baangt Execution API
====================
**baangt Execution API** provides service endpoints for executing *baangt* tests. The main features of the service are:
- running the tests
- storing test results to a database
- uoloading test run results to [*baangt Data Files*](https://gogs.earthsquad.global/athos/baangt-service/src/master/files)
- retrieving the results of the finished tests

Environmental Variables
-----------------------
You may use the following environmental variables to configure the **baangt Execution API**:

| Variable | Default Value | Description
|----------|---------------|-------------
| SECRET_KEY | secret!key | The secret key of the app
| DATABASE_URL | sqlite:///testrun.db | Database URL for *baangt* test results
| BAANGT_DATAFILE_HOST | 127.0.0.1:5050 | The host that runs [*baangt Data Files*](https://gogs.earthsquad.global/athos/baangt-service/src/master/files)
| REDIS_URL | redis:// | The Redis connection URL


Redis Worker
------------
**baangt Execution API** implements *baangt* test runs as the background tasks via Redis. So, firstly you need a running Redis server. The service uses the Redis Queue (RQ) worker named `baangt-tasks` that based on the same code to run the *baangt* tests. At least one `baangt-tasks` RQ worker should be running to support *baangt* Execution API service. Be sure to install requirements and setup environmental vars before starting the RQ workers.

```bash
pip install -r requirements.txt
export DATABASE_URL=[baangt_results_database_url]
export REDIS_URL=[your_redis_url]
rq worker baangt-tasks
``` 

Set-up
------
Before starting the **baangt Execution API** application you need just install the requirements from `requirements.txt`

```bash
pip install -r requirements.txt
```

Database
--------
**baangt Execution API** service interacts with *baangt* [Database](https://baangt.readthedocs.io/en/latest/SaveResults2Database.html)

Subclasses
----------
**baangt Execution API** service supports *BAANG* subclasses. You need just go to `plugin` folder (`api/plugin` folder in this repo) and clone (or paste) your subclasses there. For instance:
```bash
mkdir plugin
cd plugin
git clone https://gogs.earthsquad.global/athos/baangt-subclassingDemo.git
``` 
**IMPORTANT!**  
Ensure, you add the *subclasses* to [**baangt UI Web-Service**](https://gogs.earthsquad.global/athos/baangt-service/src/master/ui)

Run Service
-----------
Please use a Python WSGI Server to run the *baangt* Execution API application.  
The **application name** is `app`  
For example, to run the service with *gunicorn* use:

```bash
gunicorn -b :8000 --access-logfile access.log --error-logfile error.log app:app
```

API Reference
-------------
#### Base URL
```url
http://[HOST_NAME]/v0.1
```

#### Run Tests from XLSX
	POST [BASE_URL]/run/xlsx/

##### Requested files:
	testrun: an XLSX file containing a Testrun definition

##### Response:
```json
STATUS CODE: 202 Accepted
JSON: 
{
  "id": "[TESTRUN_UUID]"
}
```
------------------------

#### Run Tests from JSON
	POST [BASE_URL]/json

##### Requested JSON:
```json
{
  "TESTRUN_NAME": {"TESTRUN_DEFINITION_OBJECT"}
}
```
##### Response:
```json
STATUS CODE: 202 Accepted
JSON: 
{
  "id": "[TESTRUN_UUID]"
}
```
-----------------------

### Get Testrun results as JSON
	GET [BASE_URL]/results/json/[UUID]
UUID is fetched Testrun UUID as a string

##### Response:
Testrun is in progress:
```json
STATUS CODE: 202 Accepted
JSON: 
{
  "id": "[TESTRUN_UUID]"
}
``` 
Testrun is finished:
```json
STATUS CODE: 200 OK
JSON: 
{
  "id": "[TESTRUN_UUID]",
  "Name": "[TESTRUN_NAME]",
  "Summary": {"[TESTRUN_SUMMARY_OBJECT]"},
  "GlobalSettings": {"[TESTRUN_GLOBALS_OBJECT]"},
  "TestSequences": {"[TEST_SEQUANCES_OBJECT]"}
}
```

