
from tests.supports import path_to_data
import os
import json
import uuid

#
# GET home
#
def test_index(client):
    r = client.get('/')
    assert r.status_code == 200, f'status code: {r.status_code}'

def test_save(application):
    # set-up
    url = '/save'
    client = application.test_client()

    # GET method
    r = client.get(url)
    assert r.status_code == 405, f'GET method. Status code: {r.status_code}'

    # POST method without data
    r = client.post(url)
    assert r.status_code == 400, f'POST method withou data. Status code: {r.status_code}'
    # assert error in response
    response = json.loads(r.data)
    assert 'error' in response

    # POST method with wrong data
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        r = client.post(url, data={'file': f})
    assert r.status_code == 400, f'POST method with wrong data. Status code: {r.status_code}'
    # assert error in response
    response = json.loads(r.data)
    assert 'error' in response

    # POST method with correct data
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        r = client.post(url, data={'dataFile': f})
    assert r.status_code == 200
    # assert response
    response = json.loads(r.data)
    assert 'uuid' in response
    try:
        uuid.UUID(response.get('uuid'))
    except Exception as e:
        assert False, f'Exception: {e}'
    
    # assert uploaded file
    path_to_file = os.path.join(
        application.root_path,
        application.config.get('UPLOAD_FOLDER'),
        response.get('uuid'),
    )
    assert os.path.isfile(path_to_file), 'Uploaded file does NOT exists in UPLOADS'

    # remove uploaded file
    os.remove(path_to_file)

def test_update(application):
    # set-up
    file_id = uuid.uuid4()
    url = f'/update/{file_id}'
    client = application.test_client()

    # GET method
    r = client.get(url)
    assert r.status_code == 405 f'GET method. Status code: {r.status_code}'

    # POST method without data
    r = client.post(url)
    assert r.status_code == 400, f'POST method withou data. Status code: {r.status_code}'
    # assert error in response
    response = json.loads(r.data)
    assert 'error' in response

    # POST method with wrong data
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        r = client.post(url, data={'file': f})
    assert r.status_code == 400 f'POST method with wrong data. Status code: {r.status_code}'
    # assert error in response
    response = json.loads(r.data)
    assert 'error' in response

    # POST method with correct data
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        r = client.post(url, data={'dataFile': f})
    assert r.status_code == 200
    # assert response
    response = json.loads(r.data)
    assert 'uuid' in response
    assert response.get('uuid') == str(file_id), 'Response file ID does not match'
    
    # assert uploaded file
    path_to_file = os.path.join(
        application.root_path,
        application.config.get('UPLOAD_FOLDER'),
        str(file_id),
    )
    assert os.path.isfile(path_to_file), 'Uploaded file does NOT exists in UPLOADS'

    # remove uploaded file
    os.remove(path_to_file)