#!/bin/bash

# default ports
#       UI    API    DF     RD     PS
ports=("80" "5000" "5050" "6380" "5430")
# default flag for bulding docker images 
b=true

# parse arguments
args=("$@")

i=0
while [ $i -lt $# ]
do
	case ${args[$i]} in
		"-ui")
			ports[0]=${args[$i+1]}
			;;
		"-api")
			ports[1]=${args[$i+1]}
			;;
		"-df")
			ports[2]=${args[$i+1]}
			;;
		"-rd")
			ports[3]=${args[$i+1]}
			;;
		"-ps")
			ports[4]=${args[$i+1]}
			;;
		"-b")
			b=false
			let i--
			;;
		"stop")
			echo
			echo Stopping containers...
			docker stop baangt-ui baangt-api baangt-file rq-worker redis postgres
			exit 1
			;;
	esac
	let i=$i+2
done

# building docker images
if [ $b = true ]
then
	# UI Service
	echo
	echo Building UI Service...
	docker build -t baangt-ui:latest ui
	# File Servise
	echo
	echo Building DataFile Service...
	docker build -t baangt-file:latest files
	# API Service
	echo
	echo Building Execution API Service...
	docker build -t baangt-api:latest api
fi

# running containers
# Redis
echo
echo Starting Redis...
docker run -d -p ${ports[3]}:6379 --name redis --rm redis:5-alpine

# RQ Worker
echo
echo Starting Redis Worker...
docker run -d --name rq-worker  \
    -e BAANGT_DATAFILE_HOST=172.17.0.1:${ports[2]} --rm \
    -e DATABASE_URL=postgresql://baangt:12345@172.17.0.1:${ports[4]}/baangt \
    -e REDIS_URL=redis://172.17.0.1:${ports[3]}/0 --entrypoint "./runservice.rq.sh" \
    baangt-api:latest

# PostgreSQL
echo
echo Starting PostgreSQL...
docker run -d -p ${ports[4]}:5432 --name postgres -e POSTGRES_PASSWORD=12345 \
    -e POSTGRES_USER=baangt -e POSTGRES_DB=baangt --rm postgres:10.12
    
# Data Files
echo
echo Starting Data File Service...
docker run -d -p ${ports[2]}:5000 --name baangt-file --rm baangt-file:latest

# UI Service
echo
echo Starting UI Service...
docker run -d -p ${ports[0]}:5000 --name baangt-ui  \
    -e BAANGT_DATAFILE_HOST=172.17.0.1:${ports[2]} \
    -e BAANGT_API_HOST=172.17.0.1:${ports[1]} --rm baangt-ui:latest

# API Service
echo
echo Starting Execution API Service...
docker run -d -p ${ports[1]}:5000 --name baangt-api --rm \
    -e BAANGT_DATAFILE_HOST=172.17.0.1:${ports[2]}  \
    -e DATABASE_URL=postgresql://baangt:12345@172.17.0.1:${ports[4]}/baangt \
    -e REDIS_URL=redis://172.17.0.1:${ports[3]}/0 baangt-api:latest