baangt UI Web-Service
=====================
**baangt UI Web-Service** provides Web GUI for definition *baangt* tests. The main features of the service:
- creating test definitions using Web GUI
- importing test definitions from local files (XLSX, JSON formats are supported, see available [examples](https://gogs.earthsquad.global/athos/baangt-service/src/master/ui/examples))
- exporting test definitions to XLSX or JSON files
- storing test definitions in a database (see details on the database in [**Database Description**](https://gogs.earthsquad.global/athos/baangt-service/src/master/ui/docs/database.md))
- running the defined tests via [**baangt Execution API**](https://gogs.earthsquad.global/athos/baangt-service/src/master/api)

Environmental Variables
-----------------------
You may use the following environmental variables to configure the **baangt UI Web-Service**:

| Variable | Default Value | Description
|----------|---------------|-------------
| SECRET_KEY | secret!key | The secret key of the app
| DATABASE_URL | sqlite:///testrun.db | [*Database*](https://gogs.earthsquad.global/athos/baangt-service/src/master/ui/docs/database.md) URL
| BAANGT_API_HOST | 127.0.0.1:8000 | The host that runs [*baangt Execution API*](https://gogs.earthsquad.global/athos/baangt-service/src/master/api)
| BAANGT_DATAFILE_HOST | 127.0.0.1:5050 | The host that runs [*baangt Data Files*](https://gogs.earthsquad.global/athos/baangt-service/src/master/files)

Set-up
------
Before starting the **baangt UI Web-Service** application you need to define *DATABASE_URL* and run
Before starting the **baangt Data Files** application you need to:
- install the required python packages from `requirements.txt`
- define *DATABASE_URL* var
- run `flask db upgrade` to create required by the *baangt UI Web-Service* data tables 
- run `python db_update.py` to populate *baangt UI Web-Service* database with the default supporting instances

```bash
pip install -r requirements.txt
export DATABASE_URL=[your_database_url]
flask db upgrade
python db_update.py
```

Supporting Entities
-------------------
It is easily to customize the *baangt* supporting entities (namely, `classnames`, `testcases`, `browsers`, `activities`, `locators`). You may just edit file `db_supports.json` and run the update script:
```bash
python db_update.py
``` 

Subclasses
----------
To benefit subclasses that was installed in [**baangt Execution API**](https://gogs.earthsquad.global/athos/baangt-service/src/master/api) service, add them to `db_supports.json` and update the database. For instance, to introduce subclass `myTestCaseSequence.py` from [https://gogs.earthsquad.global/athos/baangt-subclassingDemo.git](https://gogs.earthsquad.global/athos/baangt-subclassingDemo.git) add it to `db_supports.json` first:
```JSON
{
	"classnames": {
		"plugin.baangt-subclassingDemo.myTestCaseSequence": "Demo subclass cloned from https://gogs.earthsquad.global/athos/baangt-subclassingDemo",
		...
	},
	...
}
```
then update the database:
```bash
python db_ipdate.py
```
**IMPORTANT!**  
Ensure, you install the new subclasses in [**baangt Execution API**](https://gogs.earthsquad.global/athos/baangt-service/src/master/api). 

Run Service
-----------
Please use a Python WSGI Server to run the *baangt* UI Web-Service application.  
The **application name** is `app`  
For example, to run the service with *gunicorn* use:

```bash
gunicorn -b :8000 --access-logfile access.log --error-logfile error.log app:app
```
