import os
import uuid
import re
from app import app, models, db
#from flask import current_app as app
from datetime import datetime
import xlsxwriter, xlrd, json, requests
from flask_login import current_user
from flask import flash
from sqlalchemy import func, desc

#
# Default VALUES
#
CLASSNAME_TESTCASESEQUENCE = 'GC.CLASSES_TESTCASESEQUENCE'
CLASSNAME_TESTCASE = 'GC.CLASSES_TESTCASE'
CLASSNAME_TESTSTEP = 'GC.CLASSES_TESTSTEPMASTER'

BROWSER_TYPE = 'FF'
TESTCASE_TYPE = 'Browser'
ACTIVITY_TYPE = 'GOTOURL'
#LOCATOR_TYPE = 'xpath'


#
# item categories
#
def getItemCategories():
	categories = {}
	categories['main'] = [
		'testrun',
		'testcase_sequence',
		'testcase',
		'teststep_sequence',
		'teststep',
	]

	return categories

#
# get name of the item_type
#
def getItemType(item_type, plural=False):	
	# main items 
	if item_type == 'testrun':
		name = 'Testrun'
	elif item_type == 'testcase_sequence':
		name = 'Test Case Sequence'
	elif item_type == 'testcase':
		name = 'Test Case'
	elif item_type == 'teststep_sequence':
		name = 'Test Step Sequence'
	elif item_type == 'teststep':
		name = 'Test Step'
	else:
		# wrong item_type
		return ''

	# check for plurals
	if plural:
		name += 's'

	return name

#
# generate choices of items
#

def getTestCaseSequences():
	return [(f'{item.uuid}', item.name) for item in models.TestCaseSequence.query.order_by(func.lower(models.TestCaseSequence.name)).all()]

def getDataFiles():
	return [(f'{item.uuid}', item.filename) for item in models.DataFile.query.order_by(func.lower(models.DataFile.filename)).all()]

def getTestCases():
	return [(f'{item.uuid}', item.name) for item in models.TestCase.query.order_by(func.lower(models.TestCase.name)).all()]

def getTestStepSequences():
	return [(f'{item.uuid}', item.name) for item in models.TestStepSequence.query.order_by(func.lower(models.TestStepSequence.name)).all()]

def getTestSteps(item_id=None):
	return [(f'{item.uuid}', item.name) for item in models.TestStepExecution.query.order_by(func.lower(models.TestStepExecution.name)).all() \
		if not item.teststep_sequence or item.teststep_sequence.uuid == item_id]

def getClassNames(default=None):
	#
	# returns:
	# default = None: choices
	# default set: id of the default instance
	#

	if default:
		item = models.ClassName.query.filter_by(name=default).first()
		if item:
			return str(item.id)
		return None

	return [(f'{item.id}', item.name) for item in models.ClassName.query.order_by(func.lower(models.ClassName.name)).all()]

def getBrowserTypes(default=None):
	#
	# returns:
	# default = None: choices
	# default set: id of the default instance
	#

	if default:
		item = models.BrowserType.query.filter_by(name=default).first()
		if item:
			return str(item.id)
		return None

	return [(f'{item.id}', item.name) for item in models.BrowserType.query.order_by(func.lower(models.BrowserType.name)).all()]

def getTestCaseTypes(default=None):
	#
	# returns:
	# default = None: choices
	# default set: id of the default instance
	#

	if default:
		item = models.TestCaseType.query.filter_by(name=default).first()
		if item:
			return str(item.id)
		return None

	return [(f'{item.id}', item.name) for item in models.TestCaseType.query.order_by(func.lower(models.TestCaseType.name)).all()]

def getActivityTypes(default=None):
	#
	# returns:
	# default = None: choices
	# default set: id of the default instance
	#

	if default:
		item = models.ActivityType.query.filter_by(name=default).first()
		if item:
			return str(item.id)
		return None

	return [(f'{item.id}', item.name) for item in models.ActivityType.query.order_by(func.lower(models.ActivityType.name)).all()]


def getLocatorTypes(default=None):
	#
	# returns:
	# default = None: choices
	# default set: id of the default instance
	#

	if default:
		item = models.LocatorType.query.filter_by(name=default).first()
		if item:
			return str(item.id)
		return None

	return [('0', 'none')] + \
		[(str(item.id), item.name) for item in models.LocatorType.query.order_by(func.lower(models.LocatorType.name)).all()]


#
# Comaprisions
#

COMPARISONS = [
	'none',
	'=',
	'>',
	'<',
	'>=',
	'<=',
	'<>',
]

def getComparisonChoices():
	return list(map(lambda x: (str(x[0]), x[1]), enumerate(COMPARISONS)))

def getComparisonId(option):
	if option:
		for index, sign in enumerate(COMPARISONS):
			if option == sign:
				return str(index)
	return '0'

#
# Get Items By Name
#
'''
def getOrCreateClassNameByName(name, description):
	# get ClassName from DB
	classname = models.ClassName.query.filter_by(name=name).first()
	if classname is None:
		# create ClassName if it doesn't exist
		classname = models.ClassName(
			name=name,
			description=description,
		)
		db.session.add(classname)
		#db.session.commit()
		app.logger.info(f'Added ClassName "{classname.name}" by {current_user}.')

	return classname 


def getBrowserTypeByName(name):
	# browser mapper
	bm = {
		'BROWSER_FIREFOX': "FF",
		'BROWSER_CHROME': "Chrome",
		'BROWSER_SAFARI': "Safari",
		'BROWSER_EDGE': "Edge",
	}
	if bm.get(name.split('.')[-1]):
		return models.BrowserType.query.filter_by(name=bm[name.split('.')[-1]]).first()
	return models.BrowserType.query.filter_by(name=name).first()

def getTestCaseTypeByName(name):
	return models.TestCaseType.query.filter_by(name=name).first()

def getActivityTypeByName(name):
	return models.ActivityType.query.filter(func.upper(models.ActivityType.name) == name.upper()).first()

def getLocatorTypeByName(name):
	if name:
		#return models.LocatorType.query.filter(func.upper(models.LocatorType.name) == name.upper()).first()
		return models.LocatorType.query.filter_by(name=name).first()
	else:
		return None
'''
def getBooleanValue(value):
	if value:
		return True
	else:
		return False


#
# Cascade Delete
#

def deleteCascade(item_type, item_id):
	#
	# implementation of cascade delete of items
	#

	# delete Testrun and its children
	if item_type == 'testrun':
		item = models.Testrun.query.get(item_id)
		# delete children calls
		for child in item.calls:
			db.session.delete(child)
			app.logger.info(f'Deleted Testrun Call {child} by {current_user}.')
			db.session.commit()
		# delete children TestCaseSequences
		for child in item.testcase_sequences:
			deleteCascade('testcase_sequence', child.id)
		# delete Testrun
		db.session.delete(item)
		db.session.commit()
		app.logger.info(f'Deleted {item_type} {item.uuid} by {current_user}.')

	# delete TestCaseSequence and its children
	elif item_type == 'testcase_sequence':
		item = models.TestCaseSequence.query.get(item_id)
		# check if item has not more then one parent
		if len(item.testrun) <= 1:
			# delete children DataFiles
			for child in item.datafiles:
				db.session.delete(child)  # TODO: delete files at DataFile Service
				app.logger.info(f'Deleted DataFile {child.uuid} by {current_user}.')
				db.session.commit()
			# delete children TestCases
			for child in item.testcases:
				deleteCascade('testcase', child.id)
			# delete TestCaseSequence
			db.session.delete(item)
			db.session.commit()
			app.logger.info(f'Deleted {item_type} {item.uuid} by {current_user}.')

	# delete TestCase and its children
	elif item_type == 'testcase':
		item = models.TestCase.query.get(item_id)
		# check if item has not more then one parent
		if len(item.testcase_sequence) <= 1:
			# delete children TestCaseSequences
			for child in item.teststep_sequences:
				deleteCascade('teststep_sequence', child.id)
			# delete TestCase
			db.session.delete(item)
			db.session.commit()
			app.logger.info(f'Deleted {item_type} {item.uuid} by {current_user}.')

	# delete TestCaseSequence and its children
	elif item_type == 'teststep_sequence':
		item = models.TestStepSequence.query.get(item_id)
		# check if item has not more then one parent
		if len(item.testcase) <= 1:
			# delete children TestStepExecutions
			for child in item.teststeps:
				deleteCascade('teststep', child.id)
			# delete TestCaseSequence
			db.session.delete(item)
			db.session.commit()
			app.logger.info(f'Deleted {item_type} {item.uuid} by {current_user}.')

	# delete TestStepExecution
	elif item_type == 'teststep':
		item = models.TestStepExecution.query.get(item_id)
		db.session.delete(item)
		db.session.commit()
		app.logger.info(f'Deleted {item_type} {item.uuid} by {current_user}.')

	# invalid type
	else:
		raise Exception(f'Item type {item_type} does not exists.')


#
# Integrity check
#

def idAsBytes(item_id):
	return uuid.UUID(item_id).bytes

def getTestrunIfCorrect(testrun_id):
	#
	# checks if Testrun tree is completed
	#

	# get testrun
	testrun = models.Testrun.query.get(idAsBytes(testrun_id))
	if not testrun:
		raise Exception(f'Testrun {testrun_id} does not exist')

	# get testrun status
	status, message = testrun.get_status()
	if not status:
		raise Exception(message)

	return testrun


#
# Testrun export/import
#

def exportJSON(testrun_id, shouldBeSaved=True):
	#
	# Export Testrun to JSON
	#

	# get testrun
	testrun = getTestrunIfCorrect(testrun_id)
	testrun_json = testrun.to_json()

	# save to file
	if shouldBeSaved:
		json_file = re.sub(r'\s+', '_', f'{testrun.name}_{datetime.now().strftime("%Y%m%d_%H%M%S")}.json')
		with open(os.path.join(app.root_path, 'static/files', json_file), 'w') as f:
			json.dump(testrun_json, f)

		return json_file

	# return dict
	else:
		return testrun_json


def exportXLSX(testrun_id):
	#
	# Exports Testrun to XLSX
	#

	# get testrun
	testrun = getTestrunIfCorrect(testrun_id)

	# create workbook
	headers = {
		'TestRun': [
			'Attribute',
			'Value',
		],

		'TestCaseSequence': [
			'Number',
			'SequenceClass',
			'TestDataFileName',
			'Sheetname',
		],
		
		'TestCase': [
			'TestCaseSequenceNumber',
			'TestCaseNumber',
			'TestCaseClass',
			'TestCaseType',
			'Browser',
		],
		
		'TestStep': [
			'TestCaseSequenceNumber',
			'TestCaseNumber',
			'TestStepNumber',
			'TestStepClass',
		],
		
		'TestStepExecution': [
			'TestCaseSequenceNumber',
			'TestCaseNumber',
			'TestStepNumber',
			'TestStepExecutionNumber',
			'Activity',
			'LocatorType',
			'Locator',
			'Value',
			'Comparison',
			'Value2',
			'Timeout',
			'Optional',
			'Release',
		],
	}

	# create workbook
	xlsx_file = re.sub(r'\s+', '_', f'{testrun.name}_{datetime.now().strftime("%Y%m%d_%H%M%S")}.xlsx')
	workbook = xlsxwriter.Workbook(os.path.join(app.root_path, 'static/files', xlsx_file))
	header_format = workbook.add_format({'bold': True})
	
	# create sheets with headers
	worksheets = {}
	for sheet, items in headers.items():
		worksheets[sheet] = workbook.add_worksheet(sheet)
		# headers
		for col, value in enumerate(items):
			# set header
			worksheets[sheet].write(0, col, value, header_format)
			# set width
			worksheets[sheet].set_column(col, col, 1.1*len(value))

	# write data
	# to TestRun
	worksheets['TestRun'].write(1, 0, 'Export Format')
	worksheets['TestRun'].write(1, 1, 'XLSX')

	# child indexes
	tc_index, tss_index, ts_index = 0, 0, 0

	# to TestCaseSequence
	for tcs_index, tcs in enumerate(testrun.testcase_sequences, 1): 
		worksheets['TestCaseSequence'].write(tcs_index, 0, tcs_index)
		worksheets['TestCaseSequence'].write(tcs_index, 1, tcs.classname.name)
		worksheets['TestCaseSequence'].write(tcs_index, 2, tcs.datafiles[0].filename)
		worksheets['TestCaseSequence'].write(tcs_index, 3, tcs.datafiles[0].sheet)

		# to TestCase
		for tc in tcs.testcases:
			tc_index += 1
			worksheets['TestCase'].write(tc_index, 0, tcs_index)
			worksheets['TestCase'].write(tc_index, 1, tc_index)
			worksheets['TestCase'].write(tc_index, 2, tc.classname.name)
			worksheets['TestCase'].write(tc_index, 3, tc.testcase_type.name)
			worksheets['TestCase'].write(tc_index, 4, tc.browser_type.name)

			# to TestStep
			for tss in tc.teststep_sequences:
				tss_index += 1
				worksheets['TestStep'].write(tss_index, 0, tcs_index)
				worksheets['TestStep'].write(tss_index, 1, tc_index)
				worksheets['TestStep'].write(tss_index, 2, tss_index)
				worksheets['TestStep'].write(tss_index, 3, tss.classname.name)

				# to TestStepExecution
				for ts in tss.teststeps:
					ts_index += 1
					worksheets['TestStepExecution'].write(ts_index, 0, tcs_index)
					worksheets['TestStepExecution'].write(ts_index, 1, tc_index)
					worksheets['TestStepExecution'].write(ts_index, 2, tss_index)
					worksheets['TestStepExecution'].write(ts_index, 3, ts_index)
					worksheets['TestStepExecution'].write(ts_index, 4, ts.activity_type.name)
					# LocatorType name
					if ts.locator_type:
						worksheets['TestStepExecution'].write(ts_index, 5, ts.locator_type.name)

					worksheets['TestStepExecution'].write(ts_index, 6, ts.locator)
					worksheets['TestStepExecution'].write(ts_index, 7, ts.value)
					worksheets['TestStepExecution'].write(ts_index, 8, ts.comparison)
					worksheets['TestStepExecution'].write(ts_index, 9, ts.value2)
					worksheets['TestStepExecution'].write(ts_index, 10, ts.timeout)
					worksheets['TestStepExecution'].write(ts_index, 11, 'X' if ts.optional else None)
					worksheets['TestStepExecution'].write(ts_index, 12, ts.release)

	workbook.close()
	return xlsx_file
'''
def uploadDataFile(file, fileId=None):
	#
	# upload DataFile to file-service
	#

	# call DataFile Service
	if fileId:
		# update
		url = '/'.join((
			'http:/',
			app.config.get('BAANGT_DATAFILE_HOST'),
			app.config.get('BAANGT_DATAFILE_UPDATE'),
			fileId,
		))
	else:
		# create
		url = '/'.join((
			'http:/',
			app.config.get('BAANGT_DATAFILE_HOST'),
			app.config.get('BAANGT_DATAFILE_SAVE'),
		))
	app.logger.info(f'Call {url} to save file {file.filename} by {current_user}.')

	r = requests.post(url, files={'dataFile': file})
	app.logger.info(f'Response to {current_user}: {r.status_code}\n{r.text}')
	return r

def importDataFile(xlsx_file, datafile_id=None):
	#
	# imports XLSX DataFile
	#

	# open xlsx
	try:
		xl = xlrd.open_workbook(file_contents=xlsx_file.read())
	except xlrd.XLRDError:
		raise Exception(f'File "{xlsx_file.filename}" could not be uploaded.')

	# get data sheet name
	sheet_name = 'data' if 'data' in xl.sheet_names() else xl.sheet_names()[0]

	# upload data file
	xlsx_file.seek(0)
	r = uploadDataFile(file=xlsx_file, fileId=datafile_id)

	# get response
	if r.status_code == 200:
		# datafile update
		if datafile_id:
			datafile = models.DataFile.query.get(uuid.UUID(datafile_id).bytes)
			if datafile:
				datafile.filename = xlsx_file.filename
				datafile.created = datetime.utcnow()
				datafile.creator = current_user
				db.session.commit()
				app.logger.info(f'Updated DataFile {datafile_id} by {current_user}.')

				return datafile_id

		# datafile create
		file_id = json.loads(r.text).get('uuid')
		datafile = models.DataFile(
			id=uuid.UUID(file_id).bytes,
			filename=xlsx_file.filename,
			sheet=sheet_name,
			creator=current_user,
		)
		db.session.add(datafile)
		db.session.commit()
		app.logger.info(f'Created DataFile {file_id} by {current_user}.')
	else:
		app.logger.error(f'Failed upload Data File "{name}" by {current_user}. {r.text}')
		raise Exception(f'Failed upload Data File "{name}". {r.text}', 'warning')

	return file_id

def importXLSX(xlsx_file, datafiles=None, item_id=None):
	#
	# imports testrun from xlsx file
	#

	# open xlsx
	try:
		xl = xlrd.open_workbook(file_contents=xlsx_file.read())
	except xlrd.XLRDError:
		raise Exception(f'File "{xlsx_file.filename}" could not be imporeted.')
	# get file name
	file_name = os.path.basename(xlsx_file.filename)

	# check for import option: create or update
	testrun = None
	if item_id is None:
		# check if imported Testrun already exists
		testrun =  models.Testrun.query.filter_by(name=file_name).first()
		if testrun is None:
			# create mode
			app.logger.info(f'Importing a Testrun from {xlsx_file.filename} by {current_user}.')
		else:
			# update mode
			item_id = testrun.id	
	else:
		# get item
		testrun = models.Testrun.query.get(item_id)
		if testrun is None: # item does not exist
			raise Exception(f'Testrun {uuid.UUID(bytes=item_id)} does not exists.')

	if testrun:
		app.logger.info(f'Updating Testrun {testrun.uuid} from {xlsx_file.filename} by {current_user}.')
	
	# Testrun object
	if testrun is None:
		# create Testrun
		testrun  = models.Testrun(
			name=file_name,
			description=f'Imported from "{file_name}"',
			creator=current_user,
		)
		db.session.add(testrun)
	else:
		# update Testrun
		testrun.description = f'Updated from "{file_name}"'
		testrun.editor = current_user
		testrun.edited = datetime.utcnow()

	# TestCaseSequence objects
	if 'TestCaseSequence' in xl.sheet_names():
		# get sheet
		testcase_sequence_sheet = xl.sheet_by_name('TestCaseSequence')
		# get headers as dict
		headers = {h.lower(): i for i, h in enumerate(testcase_sequence_sheet.row_values(0))}
		# get TestCaseSequences
		for row in range(1, testcase_sequence_sheet.nrows):
			# TCS number
			n = int(testcase_sequence_sheet.cell(row, headers['number']).value) - 1 \
				if ('number' in headers) else 0
			# ClassName
			name = testcase_sequence_sheet.cell(row, headers['sequenceclass']).value \
				if ('sequenceclass' in headers) else CLASSNAME_TESTCASESEQUENCE
			classname = getOrCreateClassNameByName(name, f'Imported from "{file_name}"')
						
			# DataFile
			datafile = None
			if headers.get('TestDataFileName'):
				# get DataFile name from sheet
				name = testcase_sequence_sheet.cell(row, headers['TestDataFileName']).value
				# check if datafile uploaded
				posted_datafile = None
				for df in datafiles:
					if df.data and df.data.filename == name:
						posted_datafile = df.data
				if posted_datafile:
					# upload Data File to File-Service
					try:
						if n < len(testrun.testcase_sequences) and testrun.testcase_sequences[n].datafiles:
							# update
							r = uploadDataFile(
								file=posted_datafile,
								fileId=testrun.testcase_sequences[n].datafiles[0].uuid,
							)
						else:
							# create
							r = uploadDataFile(file=posted_datafile)

						# get response
						if r.status_code == 200:
							file_id = json.loads(r.text).get('uuid')
							file_id_bytes = uuid.UUID(file_id).bytes
							datafile = models.DataFile.query.get(file_id_bytes)
							if datafile is None:
								# create DataFile
								datafile = models.DataFile(
									id=file_id_bytes,
									filename=file_name,
									sheet='Testcases',
									creator=current_user,
								)
								db.session.add(datafile)
						else:
							flash(f'WARNING: Failed upload Data File "{name}" for Test Case Sequence {n}. {r.text}', 'warning')
							app.logger.warning(f'Failed upload Data File "{name}" for Testrun ID {testrun.uuid} by {current_user}. {r.text}')
					
					except Exception as e:
						# failed to connect
						flash(f'WARNING: Failed upload Data File "{name}": {e}', 'warning')
						app.logger.warning(f'Failed upload Data File "{name}" for Testrun ID {testrun.uuid} by {current_user}. {e}')
				
				else:
					flash(f'WARNING: Data File "{name}" not uploaded for Test Case Sequence {n}.', 'warning')
					app.logger.warning(f'Data File "{name}" not uploaded for Testrun ID {testrun.uuid} by {current_user}.')

			else:
				flash(f'WARNING: Data File not defined for Test Case Sequence {n}.', 'warning')
				app.logger.warning(f'Data File not defined for Testrun ID {testrun.uuid} by {current_user}.')
			
			# TestCaseSequence
			if n < len(testrun.testcase_sequences):
				# update item
				testrun.testcase_sequences[n].description = f'Updated from "{file_name}"'
				testrun.testcase_sequences[n].editor = current_user
				testrun.testcase_sequences[n].edited = datetime.utcnow()
				testrun.testcase_sequences[n].classname = classname
				if datafile:
					testrun.testcase_sequences[n-1].datafiles = [datafile]
			else:
				# create item
				testcase_sequence = models.TestCaseSequence(
					name=f'{file_name}_{row}',
					description=f'Imported from "{file_name}"',
					creator=current_user,
					classname=classname,
					datafiles=[datafile] if datafile else [],
					testrun=[testrun],
				)
				db.session.add(testcase_sequence)
			
	else:
		# create default TestCaseSequence
		
		# ClassName
		# get ClassName from DB or create if it doesn't exist
		classname = getOrCreateClassNameByName(CLASSNAME_TESTCASESEQUENCE, 'Default for TestCaseSequence.')
		
		# DataFiles
		datafile = None
		if 'data' in xl.sheet_names():
			# call DataFile Service
			xlsx_file.seek(0)
			try:
				if testrun.testcase_sequences and testrun.testcase_sequences[0].datafiles:
					# update
					r = uploadDataFile(
						file=xlsx_file,
						fileId=testrun.testcase_sequences[0].datafiles[0].uuid,
					)
				else:
					# create
					r = uploadDataFile(file=xlsx_file)

				# get response
				if r.status_code == 200:
					file_id = json.loads(r.text).get('uuid')
					file_id_bytes = uuid.UUID(file_id).bytes
					datafile = models.DataFile.query.get(file_id_bytes)
					if datafile is None:
						# create DataFile
						datafile = models.DataFile(
							id=file_id_bytes,
							filename=file_name,
							creator=current_user,
						)
						db.session.add(datafile)
			except Exception as e:
				# failed to connect
				flash(f'WARNING: Failed upload Data File "{xlsx_file.filename}": {e}', 'warning')
				app.logger.warning(f'Failed upload Data File "{xlsx_file.filename}" for Testrun ID {testrun.uuid} by {current_user}. {e}')

		else:
			flash(f'WARNING: Uploaded file does not define Data.', 'warning')
			app.logger.warning(f'Uploaded file does not define Data for Testrun {testrun.uuid} by {current_user}.')

		# TestCaseSequence
		if testrun.testcase_sequences:
			# update item
			testrun.testcase_sequences[0].description = f'Updated to default.'
			testrun.testcase_sequences[0].editor = current_user
			testrun.testcase_sequences[0].edited = datetime.utcnow()
			testrun.testcase_sequences[0].classname = classname
			if datafile:
				testrun.testcase_sequences[0].datafiles = [datafile]

		else:
			# create item
			testcase_sequence = models.TestCaseSequence(
				name=f'{file_name}',
				description=f'Default for "{file_name}"',
				creator=current_user,
				classname=classname,
				datafiles=[datafile] if datafile else [],
				testrun=[testrun],
			)
			db.session.add(testcase_sequence)

	# TestCase objects
	if 'TestCase' in xl.sheet_names():
		# get sheet
		testcase_sheet = xl.sheet_by_name('TestCase')
		# get headers as dict
		headers = {h[1]: h[0] for h in enumerate(testcase_sheet.row_values(0))}
		# get TestCases
		for row in range(1, testcase_sheet.nrows):
			# get TestCaseSequenceNumber
			if headers.get('TestCaseSequenceNumber') is None:
				# default number is 0
				i = 0
			else:
				# get the number from sheet
				i = int(testcase_sheet.cell(row, headers['TestCaseSequenceNumber']).value) -1

			# get TestCaseNumber
			if headers.get('TestCaseNumber') is None:
				# default number is 0
				n = 0
			else:
				# get the number from sheet
				n = int(testcase_sheet.cell(row, headers['TestCaseNumber']).value) - 1
			
			# ClassName
			if headers.get('TestCaseClass') is None:
				# default ClassName name
				name = CLASSNAME_TESTCASE
			else:
				# get ClassName name from sheet
				name = testcase_sheet.cell(row, headers['TestCaseClass']).value
			# get ClassName from DB or create if it doesn't exist
			classname = getOrCreateClassNameByName(name, f'Imported from "{file_name}"')

			# TestCase
			# Browser Type
			if headers.get('Browser') is None:
				raise Exception('Sheet "TestCase" does not contain "Browser" column.')
			else:
				name = testcase_sheet.cell(row, headers['Browser']).value
				browser_type = getBrowserTypeByName(name)
				if browser_type is None:
					raise Exception(f'Unknown browser type "{name}": sheet "TestCase", row {row+1}.')
			# TestCase Type
			if headers.get('TestCaseType') is None:
				raise Exception('Sheet "TestCase" does not contain "TestCaseType" column.')
			else:
				name = testcase_sheet.cell(row, headers['TestCaseType']).value
				testcase_type=getTestCaseTypeByName(name)
				if testcase_type is None:
					raise Exception(f'Unknown testcase type "{name}": sheet "TestCase" row {row+1}.')

			if n < len(testrun.testcase_sequences[i].testcases):
				# update item
				testrun.testcase_sequences[i].testcases[n].description = f'Updated from "{file_name}"'
				testrun.testcase_sequences[i].testcases[n].editor = current_user
				testrun.testcase_sequences[i].testcases[n].edited = datetime.utcnow()
				testrun.testcase_sequences[i].testcases[n].classname = classname
				testrun.testcase_sequences[i].testcases[n].browser_type = browser_type
				testrun.testcase_sequences[i].testcases[n].testcase_type = testcase_type
			else:
				# create item
				testcase  = models.TestCase(
					name=f'{file_name}_{row}',
					description=f'Imported from "{file_name}"',
					creator=current_user,
					classname=classname,
					browser_type=browser_type,
					testcase_type=testcase_type,
					testcase_sequence=[testrun.testcase_sequences[i]],
				)
				db.session.add(testcase)
	else:
		# create default TestCase

		# ClassName
		# get ClassName from DB or create if it doesn't exist
		classname = getOrCreateClassNameByName(CLASSNAME_TESTCASE, 'Default for TestCase.')

		# TestCase
		if testrun.testcase_sequences[0].testcases:
			# update item
			testrun.testcase_sequences[0].testcases[0].description = f'Updated to default.'
			testrun.testcase_sequences[0].testcases[0].editor = current_user
			testrun.testcase_sequences[0].testcases[0].edited = datetime.utcnow()
			testrun.testcase_sequences[0].testcases[0].classname = classname
			testrun.testcase_sequences[0].testcases[0].browser_type = getBrowserTypeByName(BROWSER_TYPE)
			testrun.testcase_sequences[0].testcases[0].testcase_type = getTestCaseTypeByName(TESTCASE_TYPE)
		else:
			# create item
			testcase  = models.TestCase(
				name=f'{file_name}',
				description=f'Default for "{file_name}"',
				creator=current_user,
				classname=classname,
				browser_type=getBrowserTypeByName(BROWSER_TYPE),
				testcase_type=getTestCaseTypeByName(TESTCASE_TYPE),
				testcase_sequence=[testrun.testcase_sequences[0]],
			)
			db.session.add(testcase)

	# TestStepSequence objects
	if 'TestStep' in xl.sheet_names():
		# get sheet
		teststep_sheet = xl.sheet_by_name('TestStep')
		# get headers as dict
		headers = {h[1]: h[0] for h in enumerate(teststep_sheet.row_values(0))}
		# get TestSteps
		print('*** TSS')
		for row in range(1, teststep_sheet.nrows):
			# get TestStepNumber
			n = 0 # default number is 0 
			if 'TestStepNumber' in headers:
				# get the number from sheet
				n = int(teststep_sheet.cell(row, headers['TestStepNumber']).value) - 1

			# get TestCaseNumber
			j = 0 # default number is 0 
			if 'TestCaseNumber' in headers:
				# get the number from sheet
				j = int(teststep_sheet.cell(row, headers['TestCaseNumber']).value) - 1

			# get TestCaseSequenceNumber
			i = 0 # default number is 0 
			if 'TestCaseSequenceNumber' in headers:
				# get the number from sheet
				i = int(teststep_sheet.cell(row, headers['TestCaseSequenceNumber']).value) - 1

			# ClassName
			name = CLASSNAME_TESTSTEP # default ClassName name
			if 'TestStepClass' in headers:
				# get ClassName name from sheet
				name = teststep_sheet.cell(row, headers['TestStepClass']).value
			# get ClassName from DB or create if it doesn't exist
			classname = getOrCreateClassNameByName(name, f'Imported from "{file_name}"')
			print(f'{row}\t{i}\t{j}\t{n}\t{name}')
			print(f'DB\t{len(testrun.testcase_sequences)}\t{len(testrun.testcase_sequences[i].testcases)}\t{len(testrun.testcase_sequences[i].testcases[j].teststep_sequences)}')

			if n < len(testrun.testcase_sequences[i].testcases[j].teststep_sequences):
				# update item
				testrun.testcase_sequences[i].testcases[j].teststep_sequences[n].description = f'Updated from "{file_name}"'
				testrun.testcase_sequences[i].testcases[j].teststep_sequences[n].editor = current_user
				testrun.testcase_sequences[i].testcases[j].teststep_sequences[n].edited = datetime.utcnow()
				testrun.testcase_sequences[i].testcases[j].teststep_sequences[n].classname = classname
			else:
				# create item
				teststep_sequence  = models.TestStepSequence(
					name=f'{file_name}_{row}',
					description=f'Imported from "{file_name}"',
					creator=current_user,
					classname=classname,
					testcase=[testrun.testcase_sequences[i].testcases[j]],
				)
				db.session.add(teststep_sequence)
	else:
		# create default TestStep
		# ClassName
		# get ClassName from DB or create if it doesn't exist
		classname = getOrCreateClassNameByName(CLASSNAME_TESTSTEP, 'Default for TestStep')

		# TestStepSequence
		if testrun.testcase_sequences[0].testcases[0].teststep_sequences:
			# update item
			testrun.testcase_sequences[0].testcases[0].teststep_sequences[0].description = f'Updated to default.'
			testrun.testcase_sequences[0].testcases[0].teststep_sequences[0].editor = current_user
			testrun.testcase_sequences[0].testcases[0].teststep_sequences[0].edited = datetime.utcnow()
			testrun.testcase_sequences[0].testcases[0].teststep_sequences[0].classname = classname
		else:
			# create item
			teststep_sequance  = models.TestStepSequence(
				name=f'{file_name}_1',
				description=f'Default for "{file_name}"',
				creator=current_user,
				classname=classname,
				testcase=[testrun.testcase_sequences[0].testcases[0]]
			)
			db.session.add(teststep_sequance)

	# TestStepsExecution objects
	if 'TestStepExecution' in xl.sheet_names():
		# get sheet
		teststep_execution_sheet = xl.sheet_by_name('TestStepExecution')
		# get headers as dict
		headers = {h[1]: h[0] for h in enumerate(teststep_execution_sheet.row_values(0))}
		# get TestStepExecutions
		print('*** TS ROWS:')
		print(headers)
		for row in range(1, teststep_execution_sheet.nrows):
			# get TestStepExecutionNumber
			n = row - 1 # default number
			if 'TestStepExecutionNumber' in headers:
				# get the number from sheet
				n = int(teststep_execution_sheet.cell(row, headers['TestStepExecutionNumber']).value) - 1
			
			# get TestStepNumber
			k = 0
			if 'TestStepNumber' in headers:
				# get the number from sheet
				k = int(teststep_execution_sheet.cell(row, headers['TestStepNumber']).value) - 1

			# Activity Type
			if headers.get('Activity') is None:
				raise Exception('Sheet "TestStepExecution" does not contain "Activity" column.')
			else:
				name = teststep_execution_sheet.cell(row, headers['Activity']).value
				activity_type = getActivityTypeByName(name)
				if activity_type is None:
					raise Exception(f'Unknown activity type "{name}": sheet "TestStepExecution", row {row+1}')
			
			# Locator Type
			if headers.get('LocatorType') is None:
				raise Exception('Sheet "TestStepExecution" does not contain "LocatorType" column.')
			else:
				locator_type = getLocatorTypeByName(teststep_execution_sheet.cell(row, headers['LocatorType']).value)

			# get Locator
			locator = None
			if 'Locator' in headers:
				locator = teststep_execution_sheet.cell(row, headers['Locator']).value or None
			
			# get Value
			value = None
			if 'Value' in headers:
				value = teststep_execution_sheet.cell(row, headers['Value']).value or None
				if re.match(r'\d+\.0+', str(value)):
					value = int(value)
			
			# get Value 2
			value2 = None
			if 'Value2' in headers:
				value2 = teststep_execution_sheet.cell(row, headers['Value2']).value or None
				if re.match(r'\d+\.0+', str(value2)):
					value2 = int(value2)
			
			# get Comparison
			comparison = None
			if 'Comparison' in headers:
				comparison = teststep_execution_sheet.cell(row, headers['Comparison']).value or None
			
			# get Timeout
			timeout = None
			if 'Timeout' in headers:
				timeout = teststep_execution_sheet.cell(row, headers['Timeout']).value or None
			
			# get Optional
			optional = False
			if 'Optional' in headers:
				optional = getBooleanValue(teststep_execution_sheet.cell(row, headers['Optional']).value)
			
			# get Release
			release = None
			if 'Release' in headers:
				release = teststep_execution_sheet.cell(row, headers['Release']).value or None
			print(f'{row}\t{i}\t{j}\t{n}\t{name}')

			if n < len(testrun.testcase_sequences[0].testcases[0].teststep_sequences[k].teststeps):
				# update item
				teststep = testrun.testcase_sequences[0].testcases[0].teststep_sequences[k].teststeps[n]
				teststep.description = f'Updated from "{file_name}"'
				teststep.editor = current_user
				teststep.edited = datetime.utcnow()
				teststep.activity_type=activity_type
				teststep.locator_type=locator_type
				teststep.locator=locator
				teststep.value=value
				teststep.comparison=comparison
				teststep.value2=value2
				teststep.timeout=timeout
				teststep.optional=optional
				teststep.release=release
			else:
				# create item
				teststep  = models.TestStepExecution(
					name=f'{file_name}_{row}',
					description=f'Imported from "{file_name}"',
					creator=current_user,
					teststep_sequence=testrun.testcase_sequences[0].testcases[0].teststep_sequences[k],
					activity_type=activity_type,
					locator_type=locator_type,
					locator=locator,
					value=value,
					comparison=comparison,
					value2=value2,
					timeout=timeout,
					optional=optional,
					release=release,
				)
				db.session.add(teststep)

	# commit and log changes
	db.session.commit()
	item_types = getItemCategories().get('main')
	log_import(item_types[0], testrun.uuid, testrun.edited)
	for tcs in testrun.testcase_sequences:
		log_import(item_types[1], tcs.uuid, tcs.edited)
		for tc in tcs.testcases:
			log_import(item_types[2], tc.uuid, tc.edited)
			for tss in tc.teststep_sequences:
				log_import(item_types[3], tss.uuid, tss.edited)
				for ts in tss.teststeps:
					log_import(item_types[4], ts.uuid, ts.edited)


def log_import(item, item_id, updated):
	action = 'Updated' if updated else 'Created' 
	app.logger.info(f'{action} {getItemType(item)} {item_id} by {current_user}.')
'''

#
# Updating Testrun calls
#
def update_all_calls():
	for call in models.TestRunCall.query.filter_by(is_finished=False).filter_by(is_failed=False).all():
		update_call(call)

def update_call(call):
	# call API
	url = '/'.join((
		'http:/',
		app.config.get('BAANGT_API_HOST'),
		app.config.get('BAANGT_API_STATUS'),
		str(call),
	))
	app.logger.info(f'Call results by {current_user}. API URL: {url}')
	r = requests.get(url)
	app.logger.info(f'Response: {r.status_code}')

	if r.status_code == 500:
		raise Exception('API Server interanal error')
		
	elif r.status_code == 200:
		results = json.loads(r.text)
		call.is_finished = True
		call.testrecords = results['Summary'].get('TestRecords')
		call.successful = results['Summary'].get('Successful')
		call.error = results['Summary'].get('Error')
		call.paused = results['Summary'].get('Paused')
		call.stage = results['GlobalSettings'].get('Stage')
		# calculate duration in ms
		starttime = datetime.strptime(results['Summary'].get('StartTime'), '%H:%M:%S')
		endtime = datetime.strptime(results['Summary'].get('EndTime'), '%H:%M:%S')
		duration = endtime - starttime
		call.duration = duration.seconds + round(duration.microseconds/1000000)
		# commit update
		db.session.commit()
		app.logger.info(f'Testrun call {call} updated by {current_user}')
		
	elif r.status_code//100 == 4:
		call.is_failed = True
		call.error_message = r.text[-512:]
		# commit update
		db.session.commit()
		app.logger.warning(f'Testrun call {call} updated by {current_user} as failed:\n{r.text}')

