###########################################
#                                         #
#   Adds default instances to BAANGT DB   #
#                                         #
###########################################

from app import db, models
import json

path_to_defaults = 'db_supports.json'

classes = {
	"classnames": models.ClassName,
	"testcases": models.TestCaseType,
	"browsers": models.BrowserType,
	"activities": models.ActivityType,
	"locators": models.LocatorType,
}

def update_instances(cls, items):
	#
	# creates instances of class cls from dictionary items
	#

	# get existing entities
	current_set = {e.name: { 'id': e.id, 'description': e.description} for e in cls.query.all()}

	for key, value in items.items():
		# check if the item exists
		if key in current_set:
			# check for description update
			if value != current_set[key].get('description'):
				print(f"updating instance '{key}' (id: {current_set[key].get('id')})")
				instance = cls.query.get(current_set[key].get('id'))
				instance.description = value

		# create new instance
		else:
			print(f"creating instance '{key}'")
			instance = cls(name=key, description=value)
			db.session.add(instance)


if __name__ == '__main__':
	# get defaults from json
	with open(path_to_defaults, 'r') as f:
		defaults = json.load(f)

	# update supports
	for key, value in defaults.items():		
		print(f'\nUpdating {key}...')
		update_instances(classes.get(key), value)
		print('Done.')

	db.session.commit()