Database Description
====================

**baangt UI Service** supports the following databases:
- SQLite
- PostreSQL

The environmental variable DATABASE_URL shows **baangt UI Service** the location of the database.

Authentication
--------------

### Table *users*

Table holds the data on registered users.

| Column | Data Type | Description
|--------|:---------:|-----------------
| id     | INTEGER   | User's ID <br> PRIMARY KEY
| username | VARCHAR(64) | A *username* associated with the user <br> UNIQUE <br> NOT NULL
| password | VARCHAR(128) | The hashed password string associated with the user <br> NOT NULL
| created | DATETIME | The time when the user was created <br> NOT NULL
| lastlogin | DATETIME |  The time when the user last logged in <br> NOT NULL


Main entities
-------------

#### Table `testruns`

Table holds data on Test Run Definitions (TR).

| Column | Data Type | Description
|--------|:---------:|-----------------
| id     | INTEGER   | ID of the *TR* <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *TR* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *TR* <br> NOT NULL
| created | DATETIME | The time when the *TR* was created <br> NOT NULL
| creator_id | INTEGER | The *user* that created the *TR* <br> FOREIGN KEY to `users` <br> NOT NULL)
| edited | DATETIME | The time when the *TR* was last edited
| editor_id | INTEGER | The *user* that last edited the *TR* <br> FOREIGN KEY to `users`


#### Table `testcase_sequences`

Table holds data on Test Case Sequence Definitions (TCS).

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *TCS* <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *TCS* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *TCS* <br> NOT NULL
| created | DATETIME  | The time when the *TCS* was created <br> NOT NULL
| creator_id | INTEGER | The *user* that created the *TCS* <br> FOREIGN KEY to `users` <br> NOT NULL
| edited | DATETIME | The time when the *TCS* was last edited
| editor_id | INTEGER | The *user* that last edited the *TCS* <br> FOREIGN KEY to `users`
| classname_id | INTEGER | The *classname* associated with the *TCS* <br> FOREIGN KEY to `classnames` <br> NOT NULL


#### Table `datafiles`

Table holds data on baangt Data Files. 

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *DataFile* <br> PRIMARY KEY
| uuid | VARCHAR(64) | UUID string assocoated with the *DataFile* <br> NOT NULL
| filename | VARCHAR(64) | The *DataFile* name <br> NOT NULL
| sheet | VARCHAR(32) | The *sheet* name of the *DataFile* that contains the test data <br> NOT NULL
| created | DATETIME  | The time when the *DataFile* instance was created <br> NOT NULL
| creator_id | INTEGER | The *user* that created the *DataFile* instance <br> FOREIGN KEY to `users` <br> NOT NULL



#### Table `testcases`

Table holds data on Test Case Definitions (TC).

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *TC* <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *TC* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *TC* <br> NOT NULL
| created | DATETIME | The time when the *TC* was created <br> NOT NULL
| creator_id | INTEGER | The *user* that created the *TC* <br> FOREIGN KEY to `users` <br> NOT NULL
| edited | DATETIME | The time when the *TC* was last edited
| editor_id | INTEGER | The *user* that last edited the *TC* <br> FOREIGN KEY to `users`
| classname_id | INTEGER | The *classname* associated with the *TC* <br> FOREIGN KEY to `classnames` <br> NOT NULL
| browser_type_id | INTEGER | The *type of browser* that should be used to run the *TC* <br> FOREIGN KEY to `browser_types` <br> NOT NULL
| testcase_type_id | INTEGER | The *type of test case* <br> FOREIGN KEY to `testcase_types` <br> NOT NULL



#### Table `teststep_sequences`

Table holds data on Test Step Sequence Definitions (TSS).

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *TSS* <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *TSS* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *TSS* <br> NOT NULL
| created | DATETIME | The time when the *TSS* was created <br> NOT NULL
| creator_id | INTEGER | The user that created the *TSS* <br> FOREIGN KEY to `users` <br> NOT NULL
| edited | DATETIME | The time when the *TSS* was last edited
| editor_id | INTEGER | The user that last edited the *TSS* <br> FOREIGN KEY to `users`
| classname_id | INTEGER | The *classname* associated with the *TSS* <br> FOREIGN KEY to `classnames` <br> NOT NULL


#### Table `teststep_executions`

Table holds data on Definitions of Test Step Execution (TS).

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *TS* <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *TS* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *TS* <br> NOT NULL
| created | DATETIME | The time when the *TS* was created <br> NOT NULL
| creator_id | INTEGER | The user that created the *TS* <br> FOREIGN KEY to `users` <br> NOT NULL
| edited | DATETIME | The time when the *TS* was last edited
| editor_id | INTEGER | The user that last edited the *TS* <br> FOREIGN KEY to `users`
| locator | VARCHAR(512) | A string representing a *baangt* locator
| locator_type_id | INTEGER | The *locator type* used by the *TS* <br> FOREIGN KEY to `locator_types`
| optional | BOOLEAN | | Shows if the execution of the current test case should be continued after *timeout exception* was was thrown <br> NOT NULL
| timeout | DOUBLE(5,2) | Overrides the default value of the *baangt* timeout
| release | VARCHAR(64) | A release string of the *TS* 
| value | VARCHAR(1024) | A value associated with the *TS*'s activity
| value2 | VARCHAR(1024) | The second value if the *TS* activity requires a comparison
| comparison | VARCHAR(16) | A comparison string if required by the *TS*'s activity
| activity_type_id | INTEGER | The *activity type* called by the *TS* <br> FOREIGN KEY to `activity_types` <br> NOT NULL
| teststep_sequence_id | INTEGER | The *TSS* instance that contains the *TS* <br> FOREIGN KEY to `teststep_sequences`


Relationships
-------------

#### Table `testrun_casesequence`

Table holds many-to-many relationship between *TR* and *TCS*.

| Column | Data Type | Constraints
|--------|:---------:|-----------------
| testrun_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `testruns`
| testcase_sequence_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `testcase_sequences`


#### Table `testcase_sequence_datafile`

Table holds many-to-many relationship between *TCS* and *DataFiles*.

| Column | Data Type | Constraints
|--------|:---------:|-----------------
| testcase_sequence_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `testcase_sequences`
| datafile_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `datafiles`


#### Table `testcase_sequence_case`

Table holds many-to-many relationship between *TCS* and *TC*.

| Column | Data Type | Constraints
|--------|:---------:|-----------------
| testcase_sequence_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `testcase_sequences`
| testcase_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `testcases`


#### Table `testcase_stepsequence`

Table holds many-to-many relationship between *TC* and *TSS*.

| Column | Data Type | Constraints
|--------|:---------:|-----------------
| testcase_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `testcases`
| teststep_sequence_id | INTEGER | PRIMARY KEY <br> FOREIGN KEY to `teststep_sequences`



Supporting entities
-------------------

#### Table `global_teststep_executions`

Table holds data on Definitions of Global Test Step Execution (GTS).

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *GTS* <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *GTS* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *GTS* <br> NOT NULL
| created | DATETIME | The time when the *GTS* was created <br> NOT NULL
| creator_id | INTEGER | The user that created the *GTS* <br> FOREIGN KEY to `users` <br> NOT NULL
| edited | DATETIME | The time when the *GTS* was last edited
| editor_id | INTEGER | The user that last edited the *GTS* <br> FOREIGN KEY to `users`
| locator_type_id | INTEGER | The *locator type* used by the *GTS* <br> FOREIGN KEY to `locator_types`
| activity_type_id | INTEGER | The *activity type* called by the *GTS* <br> FOREIGN KEY to `activity_types` <br> NOT NULL
| teststep_sequence_id | INTEGER | The *TSS* instance that contains the *GTS* <br> FOREIGN KEY to `teststep_sequences`


#### Table `classnames`

Table holds data on the *baangt* Class Names.

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *ClassName*, <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *ClassName* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *ClassName* <br> NOT NULL

#### Table `browser_types`

Table holds data on the Browser Types.

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *BrowserType*, <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *BrowserType* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *BrowserType* <br> NOT NULL

#### Table `testcase_types`

Table holds data on the Types of Test Cases (TCT).

| Column | Data Type | Description
|--------|:---------:|----------------
| id | INTEGER | ID of the *TCT*, <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *TCT* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *TCT* <br> NOT NULL

#### Table `activity_types`

Table holds data on the Activity Types of *baangt* tests.

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *ActivityType*, <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *ActivityType* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *ActivityType* <br> NOT NULL

#### Table `locator_types`

Table holds data on the Locator Types of *baangt* test activity.

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | INTEGER | ID of the *LocatorType*, <br> PRIMARY KEY
| name | VARCHAR(64) | A name associated with the *LocatorType* <br> NOT NULL
| description | VARCHAR(512) | A brief description of the *LocatorType* <br> NOT NULL


Results
-------

#### Table `testrun_calls`

Table holds data on executed Testruns.

| Column | Data Type | Description
|--------|:---------:|-----------------
| id | BLOB(16) | UUID  of the *Testrun execution* as a binary string <br> PRIMARY KEY
| created | DATETIME | The time when the *Testrun execution* was called <br> NOT NULL
| creator_id | INTEGER | The user that called the *Testrun execution* <br> FOREIGN KEY to `users` <br> NOT NULL
| testrun_id | INTEGER | The *Testrun* that was executed <br> FOREIGN KEY to `testruns` <br> NOT NULL