import pytest
from tests.supports import login
from app import models
from lxml import html
from random import choice
import uuid


#
# Cascade delete
#
@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_testrun_cascade(client, test_user, populate):
    # delete testrun item and its children
    item = choice(models.Testrun.query.all())
    url = f'/testrun/{item.uuid}/delete'

    # unauthorized access
    r = client.post(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.post(url)
    assert r.status_code == 200, f'Deleting Testrun. Status code: {r.status_code}'
    item_num = populate[0]-1
    assert models.Testrun.query.count() == item_num, 'Testrun NOT deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence NOT deleted'
    assert models.DataFile.query.count() == item_num, 'Data File NOT deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case NOT deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence NOT deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step NOT deleted'

    # try to delete item by random id
    url = f'/testrun/{uuid.uuid4()}/delete'
    r = client.post(url)
    assert r.status_code == 400, f'Deleting item that does not exist. Status code: {r.status_code}'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_testcase_sequence_cascade(client, test_user, populate):
    # delete testcase sequence item and its children
    item = choice(models.TestCaseSequence.query.all())
    url = f'/testcase_sequence/{item.uuid}/delete'

    # unauthorized access
    r = client.post(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.post(url)
    assert r.status_code == 200, f'Deleting TestCaseSequence. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num = populate[0]*populate[1] - 1
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence NOT deleted'
    assert models.DataFile.query.count() == item_num, 'Data File NOT deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case NOT deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence NOT deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step NOT deleted'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_testcase_cascade(client, test_user, populate):
    # delete testcase item and its children
    item = choice(models.TestCase.query.all())
    url = f'/testcase/{item.uuid}/delete'

    # unauthorized access
    r = client.post(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.post(url)
    assert r.status_code == 200, f'Deleting TestCase. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num = item_num*populate[1] - 1
    assert models.TestCase.query.count() == item_num, 'Test Case NOT deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence NOT deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step NOT deleted'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_teststep_sequence_cascade(client, test_user, populate):
    # delete teststep sequence item and its children
    item = choice(models.TestStepSequence.query.all())
    url = f'/teststep_sequence/{item.uuid}/delete'

    # unauthorized access
    r = client.post(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.post(url)
    assert r.status_code == 200, f'Deleting TestStepSequence. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case deleted'
    item_num = item_num*populate[1] - 1
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence NOT deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step NOT deleted'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_teststep_cascade(client, test_user, populate):
    # delete teststep item
    item = choice(models.TestStepExecution.query.all())
    url = f'/teststep/{item.uuid}/delete'

    # unauthorized access
    r = client.post(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.post(url)
    assert r.status_code == 200, f'Deleting TestStep. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence deleted'
    item_num = item_num*populate[1] - 1 + (3 if populate[0] == 1 else 0) # extra for 1 testrun free steps
    assert models.TestStepExecution.query.count() == item_num, 'Test Step NOT deleted'


#
# Delete only the item
#
@pytest.mark.parametrize('count', [(1,1), (3,2)])
def test_delete_testrun(client, test_user, populate):
    # delete testrun item
    item = choice(models.Testrun.query.all())
    url = f'/testrun/{item.uuid}/delete'

    # unauthorized access
    r = client.delete(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.delete(url)
    assert r.status_code == 200, f'Deleting Testrun. Status code: {r.status_code}'
    item_num = populate[0]-1
    assert models.Testrun.query.count() == item_num, 'Testrun NOT deleted'
    item_num = populate[0]*populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step deleted'

    # try to delete item by random id
    url = f'/testrun/{uuid.uuid4()}/delete'
    r = client.delete(url)
    assert r.status_code == 400, f'Deleting item that does not exist. Status code: {r.status_code}'

@pytest.mark.parametrize('count', [(1,1), (3,2)])
def test_delete_testcase_sequence(client, test_user, populate):
    # delete testcase sequence item
    item = choice(models.TestCaseSequence.query.all())
    url = f'/testcase_sequence/{item.uuid}/delete'

    # unauthorized access
    r = client.delete(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.delete(url)
    assert r.status_code == 200, f'Deleting TestCaseSequence. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num-1, 'Test Case Sequence NOT deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step deleted'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_testcase(client, test_user, populate):
    # delete testcase item
    item = choice(models.TestCase.query.all())
    url = f'/testcase/{item.uuid}/delete'
    
    # unauthorized access
    r = client.delete(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.delete(url)
    assert r.status_code == 200, f'Deleting TestCase. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num-1, 'Test Case NOT deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step deleted'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_teststep_sequence(client, test_user, populate):
    # delete teststep sequence item and its children
    item = choice(models.TestStepSequence.query.all())
    url = f'/teststep_sequence/{item.uuid}/delete'

    # unauthorized access
    r = client.delete(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.delete(url)
    assert r.status_code == 200, f'Deleting TestStepSequence. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num-1, 'Test Step Sequence NOT deleted'
    item_num = item_num*populate[1] + (3 if populate[0] == 1 else 0) # extra free steps for 1 testrun
    assert models.TestStepExecution.query.count() == item_num, 'Test Step deleted'

@pytest.mark.parametrize('count', [(1,1), (5,3)])
def test_delete_teststep(client, test_user, populate):
    # delete teststep item
    item = choice(models.TestStepExecution.query.all())
    url = f'/teststep/{item.uuid}/delete'

    # unauthorized access
    r = client.delete(url)
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    # assert redirection url
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # authorized access
    login(client, test_user)
    r = client.delete(url)
    assert r.status_code == 200, f'Deleting TestStep. Status code: {r.status_code}'
    item_num = populate[0]
    assert models.Testrun.query.count() == item_num, 'Testrun deleted'
    item_num *= populate[1]
    assert models.TestCaseSequence.query.count() == item_num, 'Test Case Sequence deleted'
    assert models.DataFile.query.count() == item_num, 'Data File deleted'
    item_num *= populate[1]
    assert models.TestCase.query.count() == item_num, 'Test Case deleted'
    item_num *= populate[1]
    assert models.TestStepSequence.query.count() == item_num, 'Test Step Sequence deleted'
    item_num = item_num*populate[1] - 1 + (3 if populate[0] == 1 else 0) # extra for 1 testrun free steps
    assert models.TestStepExecution.query.count() == item_num, 'Test Step NOT deleted'

#
# Delete fake item
#
def test_delete_fake_item(client, test_user):
    # set-up
    fake_item_id = uuid.uuid4()
    url = f'/fakeitem/{fake_item_id}/delete'

    # authorized access
    login(client, test_user)

    # POST method
    r = client.post(url)
    assert r.status_code == 400, f'Deleting fake item. Status code: {r.status_code}'

    # DELETE method
    r = client.delete(url)
    assert r.status_code == 400, f'Deleting fake item. Status code: {r.status_code}'


#
# GET method
#
def test_get_method(client, test_user):
    # set-up
    item_id = uuid.uuid4()
    url = f'/testrun/{item_id}/delete'

    # authorized access
    login(client, test_user)

    # GET method
    r = client.get(url)
    assert r.status_code == 405, f'GET method. Status code: {r.status_code}'