from app import models
from tests.supports import path_to_data, item_data, get_datafile_url
from datetime import datetime, timedelta
from os import path
from random import choice
import json
import uuid
import pytest


#
# User model tests
#
def test_user(session, test_user):
    # get test_user from db
    users = models.User.query.filter_by(username=test_user.get('username')).all()
    assert len(users) == 1, f'There are {len(users)} test users'
    
    # assert properties
    user = users[0]
    assert user.id == 1, f'User id : {user.id}. Should be 1'
    assert user.username == test_user.get('username'), 'Username does not match'
    assert user.password != test_user.get('password'), 'Password is not hashed'
    assert type(user.created) == type(datetime.now()), f"Created is {type(user.created)}. Should be 'datetime'"
    assert type(user.lastlogin) == type(datetime.now()), f"Lastlogin is {type(user.lastlogin)}. Should be 'datetime'"

    # assert methods
    assert str(user) == test_user.get('username'), f'__str__ returns: {str(user)}. Should be username'
    assert user.verify_password(test_user.get('password')), 'Password does not match'


def test_duplicate_user(session, test_user):
    # create duplicated user
    user = models.User(username=test_user.get('username'))
    user.set_password('Te$tPa$Sw0rD')
    session.add(user)
    
    try:
        session.commit()
        assert False, 'Duplicated user created. Should be forbidden'
    except Exception:
        # test is successful
        assert True

def test_nopassword_user(session):
    # create user
    user = models.User(username='test')
    session.add(user)
    
    try:
        session.commit()
        assert False, 'Created User without password. Should be forbidden'
    except Exception:
        # test is successful
        assert True 


#
# ClassName model tests
#
def test_classname(session):
    # create test ClassName entity
    classname = models.ClassName(
        name=item_data.get('name'),
        description=item_data.get('description'),
    )
    session.add(classname)
    session.commit()

    # get test ClassName from db
    items = models.ClassName.query.filter_by(name=classname.name).all()
    assert len(items) == 1, f'There are {len(items)} test ClassNames'
    
    # assert properties
    item = items[0]
    assert item.id == 1, f'ClassName id : {item.id}. Should be 1'
    assert item.name == item_data.get('name'), 'Name does not match'
    assert item.description == item_data.get('description'), 'Description does not match'

    # assert methods
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'


#
# BrowserType model tests
#
def test_browser(session):
    # create test Browser Type entity
    browser = models.BrowserType(
        name=item_data.get('name'),
        description=item_data.get('description'),
    )
    session.add(browser)
    session.commit()

    # get test browser from db
    items = models.BrowserType.query.filter_by(name=browser.name).all()
    assert len(items) == 1, f'There are {len(items)} test Browsers'
    
    # assert properties
    item = items[0]
    assert item.id == 1, f'Browser id : {item.id}. Should be 1'
    assert item.name == item_data.get('name'), 'Name does not match'
    assert item.description == item_data.get('description'), 'Description does not match'

    # assert methods
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'


#
# TestCaseType model tests
#
def test_testcase_type(session):
    # create test TestCase Type entity
    testcase = models.TestCaseType(
        name=item_data.get('name'),
        description=item_data.get('description'),
    )
    session.add(testcase)
    session.commit()

    # get test testcase type from db
    items = models.TestCaseType.query.filter_by(name=testcase.name).all()
    assert len(items) == 1, f'There are {len(items)} test TestCase types'
    
    # assert properties
    item = items[0]
    assert item.id == 1, f'TestCaseType id : {item.id}. Should be 1'
    assert item.name == item_data.get('name'), 'Name does not match'
    assert item.description == item_data.get('description'), 'Description does not match'

    # assert methods
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'


#
# ActivityType model tests
#
def test_activity(session):
    # create test Activity Type entity
    activity = models.ActivityType(
        name=item_data.get('name'),
        description=item_data.get('description'),
    )
    session.add(activity)
    session.commit()

    # get test activity from db
    items = models.ActivityType.query.filter_by(name=activity.name).all()
    assert len(items) == 1, f'There are {len(items)} test Activity Types'
    
    # assert properties
    item = items[0]
    assert item.id == 1, f'Activity id : {item.id}. Should be 1'
    assert item.name == item_data.get('name'), 'Name does not match'
    assert item.description == item_data.get('description'), 'Description does not match'

    # assert methods
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'


#
# LocatorType model tests
#
def test_locator(session):
    # create test Locator Type entity
    locator = models.LocatorType(
        name=item_data.get('name'),
        description=item_data.get('description'),
    )
    session.add(locator)
    session.commit()

    # get test locator from db
    items = models.LocatorType.query.filter_by(name=locator.name).all()
    assert len(items) == 1, f'There are {len(items)} test locators'
    
    # assert properties
    item = items[0]
    assert item.id == 1, f'Locator id : {item.id}. Should be 1'
    assert item.name == item_data.get('name'), 'Name does not match'
    assert item.description == item_data.get('description'), 'Description does not match'

    # assert methods
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'


#
# Testrun model tests
#
def test_testrun(session, test_user):
    # get user
    user = models.User.query.filter_by(username=test_user.get('username')).first()

    # check back references before creating item
    assert len(user.created_testruns) == 0, 'Test user already created Testruns'
    assert len(user.edited_testruns) == 0, 'Test user already edited Testruns'

    # add item
    with open(path.join(path_to_data, 'testrun.json'), 'r') as f:
        test_item = json.load(f)
    testrun = models.Testrun(
        name=test_item.get('name'),
        description=item_data.get('description'),
        creator=user,
    )
    session.add(testrun)
    session.commit()

    # get items from db
    items = models.Testrun.query.filter_by(name=test_item.get('name')).all()
    assert len(items) == 1, f'There are {len(items)} Testruns'
    
    # assert properties 
    item = items[0]
    assert type(item.id) == type(b''), 'Item id should be binary'
    assert item.name == test_item.get('name'), f'{item.name} != {test_item.get("name")}: name does not match'
    assert item.description == item_data.get('description'), f'{item.description} != {item_data.get("description")}: description does not match'
    assert type(item.created) == type(datetime.now()), f"created: {type(item.created)}. Should be of type 'datetime'"
    assert item.creator == user, f'Created by {item.creator}: should be {user}'
    assert item.edited is None , f'Edited {item.edited}: should be empty'
    assert item.editor is None, f'Editor {item.editor}: should be empty'
    assert item.global_settings is None, 'Global settings should be empty'

    # assert back references
    assert user.created_testruns == [item]
    assert len(user.edited_testruns) == 0, 'Test user already edited Testruns'

    # assert methods
    assert type(item.uuid) == type(''), 'UIID should be a string'
    assert str(item) == test_item.get('name'), f'__str__ returns: {str(item)}. Should be item name'
    assert len(item.finished_calls()) == 0, 'Item should not have finished calls'
    assert item.to_json() == test_item, 'Generated wrong JSON'

def test_testrun_edit(session, test_user):
    # get user
    user = models.User.query.filter_by(username=test_user.get('username')).first()

    # add item
    with open(path.join(path_to_data, 'testrun.json'), 'r') as f:
        test_item = json.load(f)
    item = models.Testrun(
        name=test_item.get('name'),
        description=item_data.get('description'),
        creator=user,
    )
    session.add(item)
    session.commit()

    # update item
    item.editor = user
    item.edited = datetime.utcnow()
    session.commit()

    # assert properties
    assert type(item.edited) == type(datetime.now()), f"edited: {type(item.edited)}. Should be of type 'datetime'"
    assert item.editor == user, f'Edited by {item.editor}: should be {user}'

    # assert back references
    assert user.edited_testruns == [item]


#
# TestCaseSequence model tests
#
def test_testcase_sequence(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and classname
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    classname = models.ClassName.query.first()

    # check back references before creating item
    assert len(user.created_testcase_sequences) == 0, 'Test user already created TestCaseSequences'
    assert len(user.edited_testcase_sequences) == 0, 'Test user already edited TestCaseSequences'

    # add item
    with open(path.join(path_to_data, 'testcase_sequence.json'), 'r') as f:
        test_item = json.load(f)
    test_item['SequenceClass'] = classname.name
    testcase_sequence = models.TestCaseSequence(
        name=item_data.get('name'),
        description=item_data.get('description'),
        classname=classname,
        creator=user,
    )
    session.add(testcase_sequence)
    session.commit()

    # get items from db
    items = models.TestCaseSequence.query.filter_by(name=item_data.get('name')).all()
    assert len(items) == 1, f'There are {len(items)} TestCaseSequences'
    
    # assert properties 
    item = items[0]
    assert type(item.id) == type(b''), 'Item id should be binary'
    assert item.name == item_data.get('name'), f'{item.name} != {item_data.get("name")}: name does not match'
    assert item.description == item_data.get('description'), f'{item.description} != {item_data.get("description")}: description does not match'
    assert type(item.created) == type(datetime.now()), f"created: {type(item.created)}. Should be of type 'datetime'"
    assert item.creator == user, f'Created by {item.creator}: should be {user}'
    assert item.edited is None , f'Edited {item.edited}: should be empty'
    assert item.editor is None, f'Editor {item.editor}: should be empty'
    assert item.classname == classname, f'Classname: {item.classname}. Should be {classname}'
    assert len(item.testrun) == 0, f'Testruns {item.testrun}. Should be empty'

    # assert back references
    assert user.created_testcase_sequences == [item]
    assert len(user.edited_testcase_sequences) == 0, 'Test user already edited TestCaseSequences'
    assert classname.testcase_sequences == [item]

    # assert methods
    assert type(item.uuid) == type(''), 'UIID should be a string'
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'
    assert item.to_json() == [classname.name, test_item], 'Generated wrong JSON'

def test_testcase_sequence_edit(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and classname
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    classname = models.ClassName.query.first()

    # add item
    with open(path.join(path_to_data, 'testcase_sequence.json'), 'r') as f:
        test_item = json.load(f)
    #test_item['SequenceClass'] = classname.name
    item = models.TestCaseSequence(
        name=item_data.get('name'),
        description=item_data.get('description'),
        classname=classname,
        creator=user,
    )
    session.add(item)
    session.commit()

    # update item
    item.editor = user
    item.edited = datetime.utcnow()
    session.commit()

    # assert properties
    assert type(item.edited) == type(datetime.now()), f"Edited: {type(item.edited)}. Should be of type 'datetime'"
    assert item.editor == user, f'Edited by {item.editor}: should be {user}'

    # assert back references
    assert user.edited_testcase_sequences == [item]


#
# DataFile model tests
#
def test_datafile(application, session, test_user):
    # create test support entities
    #db_supports(1)

    # get user and classname
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    #classname = models.ClassName.query.first()

    # check back references before creating item
    assert len(user.created_datafiles) == 0, 'Test user already created DataFiles'

    # add item
    datafile = models.DataFile(
        filename=item_data.get('name'),
        sheet=item_data.get('description'),
        creator=user,
    )
    session.add(datafile)
    session.commit()

    # get items from db
    items = models.DataFile.query.filter_by(filename=item_data.get('name')).all()
    assert len(items) == 1, f'There are {len(items)} DataFiles'
    
    # assert properties 
    item = items[0]
    assert type(item.id) == type(b''), 'Item id should be binary'
    assert item.filename == item_data.get('name'), f'{item.filename} != {item_data.get("name")}: filename does not match'
    assert item.sheet == item_data.get('description'), f'{item.sheet} != {item_data.get("description")}: sheet name does not match'
    assert type(item.created) == type(datetime.now()), f"created: {type(item.created)}. Should be of type 'datetime'"
    assert item.creator == user, f'Created by {item.creator}: should be {user}'
    assert len(item.testcase_sequence) == 0, f'TestCaseSequences {item.testcase_sequence}: should be empty'

    # assert back references
    assert user.created_datafiles == [item]

    # assert methods
    assert type(item.uuid) == type(''), 'UIID should be a string'
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'
    assert item.url == get_datafile_url(application.config, item.uuid), f'Wrong DataFile URL'


#
# TestCase model tests
#
def test_testcase(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and supports
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    classname = models.ClassName.query.first()
    browser = models.BrowserType.query.first()
    testcase_type = models.TestCaseType.query.first()

    # check back references before creating item
    assert len(user.created_testcases) == 0, 'Test user already created TestCases'
    assert len(user.edited_testcases) == 0, 'Test user already edited TestCases'

    # add item
    testcase = models.TestCase(
        name=item_data.get('name'),
        description=item_data.get('description'),
        classname=classname,
        browser_type=browser,
        testcase_type=testcase_type,
        creator=user,
    )
    session.add(testcase)
    session.commit()

    # get items from db
    items = models.TestCase.query.filter_by(name=item_data.get('name')).all()
    assert len(items) == 1, f'There are {len(items)} TestCases'
    
    # assert properties 
    item = items[0]
    assert type(item.id) == type(b''), 'Item id should be binary'
    assert item.name == item_data.get('name'), f'{item.name} != {item_data.get("name")}: name does not match'
    assert item.description == item_data.get('description'), f'{item.description} != {item_data.get("description")}: description does not match'
    assert type(item.created) == type(datetime.now()), f"created: {type(item.created)}. Should be of type 'datetime'"
    assert item.creator == user, f'Created by {item.creator}: should be {user}'
    assert item.edited is None , f'Edited {item.edited}: should be empty'
    assert item.editor is None, f'Editor {item.editor}: should be empty'
    assert item.classname == classname, f'Classname: {item.classname}. Should be {classname}'
    assert item.browser_type == browser, f'Browser: {item.browser_type}. Should be {browser}'
    assert item.testcase_type == testcase_type, f'Testcase type: {item.testcase_type}. Should be {testcase_type}'
    assert len(item.testcase_sequence) == 0, f'TestCaseSequences {item.testcase_sequence}: should be empty'

    # assert back references
    assert user.created_testcases == [item]
    assert len(user.edited_testcases) == 0, 'Test user already edited TestCases'
    assert classname.testcases == [item]
    assert browser.testcases == [item]
    assert testcase_type.testcases == [item]

    # assert methods
    assert type(item.uuid) == type(''), 'UIID should be a string'
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'
    
    testcase_json = [
        classname.name,
        {
            'GC.KWARGS_TESTCASETYPE': testcase_type.name,
            'GC.KWARGS_BROWSER': browser.name,
            'GC.BROWSER_ATTRIBUTES': '',
        },
        {
            'GC.STRUCTURE_TESTSTEP': {},
        }
    ]
    assert item.to_json() == testcase_json, 'Generated wrong JSON'

def test_testcase_edit(session, test_user, db_supports):
     # create test support entities
    db_supports(1)

    # get user and supports
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    classname = models.ClassName.query.first()
    browser = models.BrowserType.query.first()
    testcase_type = models.TestCaseType.query.first()

    # add item
    item = models.TestCase(
        name=item_data.get('name'),
        description=item_data.get('description'),
        classname=classname,
        browser_type=browser,
        testcase_type=testcase_type,
        creator=user,
    )
    session.add(item)
    session.commit()

    # update item
    item.editor = user
    item.edited = datetime.utcnow()
    session.commit()

    # assert properties
    assert type(item.edited) == type(datetime.now()), f"Edited: {type(item.edited)}. Should be of type 'datetime'"
    assert item.editor == user, f'Edited by {item.editor}: should be {user}'

    # assert back references
    assert user.edited_testcases == [item]


#
# TestStepSequence model tests
#
def test_teststep_sequence(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and supports
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    classname = models.ClassName.query.first()

    # check back references before creating item
    assert len(user.created_teststep_sequences) == 0, 'Test user already created TestStepSequences'
    assert len(user.edited_teststep_sequences) == 0, 'Test user already edited TestStepSequences'

    # add item
    teststep_sequence = models.TestStepSequence(
        name=item_data.get('name'),
        description=item_data.get('description'),
        classname=classname,
        creator=user,
    )
    session.add(teststep_sequence)
    session.commit()

    # get items from db
    items = models.TestStepSequence.query.filter_by(name=item_data.get('name')).all()
    assert len(items) == 1, f'There are {len(items)} TestStepSequences'
    
    # assert properties 
    item = items[0]
    assert type(item.id) == type(b''), 'Item id should be binary'
    assert item.name == item_data.get('name'), f'{item.name} != {item_data.get("name")}: name does not match'
    assert item.description == item_data.get('description'), f'{item.description} != {item_data.get("description")}: description does not match'
    assert type(item.created) == type(datetime.now()), f"created: {type(item.created)}. Should be of type 'datetime'"
    assert item.creator == user, f'Created by {item.creator}: should be {user}'
    assert item.edited is None , f'Edited {item.edited}: should be empty'
    assert item.editor is None, f'Editor {item.editor}: should be empty'
    assert item.classname == classname, f'Classname: {item.classname}. Should be {classname}'
    assert len(item.testcase) == 0, f'TestCases {item.testcase_sequence}: should be empty'

    # assert back references
    assert user.created_teststep_sequences == [item]
    assert len(user.edited_teststep_sequences) == 0, 'Test user already edited TestStepSequence'
    assert classname.teststep_sequences == [item]

    # assert methods
    assert type(item.uuid) == type(''), 'UIID should be a string'
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'
    
    teststep_sequence_json = [
        {
            'TestStepClass': classname.name,
        },
        {
            'GC.STRUCTURE_TESTSTEPEXECUTION': {},
        }
    ]
    assert item.to_json() == teststep_sequence_json, 'Wrong JSON generated'

def test_teststep_sequence_edit(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and supports
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    classname = models.ClassName.query.first()

    # add item
    item = models.TestStepSequence(
        name=item_data.get('name'),
        description=item_data.get('description'),
        classname=classname,
        creator=user,
    )
    session.add(item)
    session.commit()

    # update item
    item.editor = user
    item.edited = datetime.utcnow()
    session.commit()

    # assert properties
    assert type(item.edited) == type(datetime.now()), f"Edited: {type(item.edited)}. Should be of type 'datetime'"
    assert item.editor == user, f'Edited by {item.editor}: should be {user}'

    # assert back references
    assert user.edited_teststep_sequences == [item]


#
# TestStepExecution model tests
#
def test_teststep(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and supports
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    activity = models.ActivityType.query.first()

    # check back references before creating item
    assert len(user.created_teststeps) == 0, 'Test user already created TestSteps'
    assert len(user.edited_teststeps) == 0, 'Test user already edited TestSteps'

    # add item
    with open(path.join(path_to_data, 'teststep.json'), 'r') as f:
        test_item = json.load(f)
    test_item['Activity'] = activity.name
    teststep = models.TestStepExecution(
        name=item_data.get('name'),
        description=item_data.get('description'),
        activity_type=activity,
        creator=user,
    )
    session.add(teststep)
    session.commit()

    # get items from db
    items = models.TestStepExecution.query.filter_by(name=item_data.get('name')).all()
    assert len(items) == 1, f'There are {len(items)} TestSteps'
    
    # assert properties 
    item = items[0]
    assert type(item.id) == type(b''), 'Item id should be binary'
    assert item.name == item_data.get('name'), f'{item.name} != {item_data.get("name")}: name does not match'
    assert item.description == item_data.get('description'), f'{item.description} != {item_data.get("description")}: description does not match'
    assert type(item.created) == type(datetime.now()), f"created: {type(item.created)}. Should be of type 'datetime'"
    assert item.creator == user, f'Created by {item.creator}: should be {user}'
    assert item.edited is None , f'Edited {item.edited}: should be empty'
    assert item.editor is None, f'Editor {item.editor}: should be empty'
    assert not item.optional, 'Optional: default value should be False'
    assert item.activity_type == activity, f'Activity: {item.activity_type}. Should be {activity}'
    assert item.teststep_sequence is None, f'TestStepSequence {item.teststep_sequence}: should be empty'

    # assert back references
    assert user.created_teststeps == [item]
    assert len(user.edited_teststeps) == 0, 'Test user already edited TestStep'
    assert activity.teststeps == [item]

    # assert methods
    assert type(item.uuid) == type(''), 'UIID should be a string'
    assert str(item) == item_data.get('name'), f'__str__ returns: {str(item)}. Should be item name'
    assert item.to_json() == test_item, 'Generated wrong JSON'

def test_teststep_edit(session, test_user, db_supports):
    # create test support entities
    db_supports(1)

    # get user and supports
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    activity = models.ActivityType.query.first()

    # add item
    item = models.TestStepExecution(
        name=item_data.get('name'),
        description=item_data.get('description'),
        activity_type=activity,
        creator=user,
    )
    session.add(item)
    session.commit()

    # update item
    item.editor = user
    item.edited = datetime.utcnow()
    session.commit()

    # assert properties
    assert type(item.edited) == type(datetime.now()), f"Edited: {type(item.edited)}. Should be of type 'datetime'"
    assert item.editor == user, f'Edited by {item.editor}: should be {user}'

    # assert back references
    assert user.edited_teststeps == [item]

#
# Model's status method
#
@pytest.mark.parametrize('count', [(1,1), (5,2)])
def test_status_testrun(session, test_user, populate):
    # get random testrun
    item = choice(models.Testrun.query.all())

    status, msg = item.get_status()
    assert status
    assert msg == 'OK'

    # remove datafiles
    datafile = item.testcase_sequences[0].datafiles[0]
    item.testcase_sequences[0].datafiles = []
    session.commit()
    status, msg = item.get_status()
    assert not status
    assert 'not contain DataFiles' in msg

    # remove steps of a testcase sequence
    item.testcase_sequences[0].datafiles = [datafile]
    item.testcase_sequences[0].testcases[0].teststep_sequences[0].teststeps = []
    session.commit()
    status, msg = item.get_status()
    assert not status
    assert 'not contain TestSteps' in msg

    # remove step sequences of a testcase sequence
    item.testcase_sequences[0].testcases[0].teststep_sequences = []
    session.commit()
    status, msg = item.get_status()
    assert not status
    assert 'not contain TestStepSequences' in msg

    # remove testcases of a testcase sequence
    item.testcase_sequences[0].testcases = []
    session.commit()
    status, msg = item.get_status()
    assert not status
    assert 'not contain TestCases' in msg


#
# TestRun Call
#
@pytest.mark.parametrize('count', [(1,1)])
def test_testrun_call(session, test_user, populate):
    # get user and testrun
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    testrun = models.Testrun.query.first()

    # check back references before creating item
    assert len(user.testruns_called) == 0, 'Test user already ran Testruns'
    assert len(testrun.calls) == 0, 'Test Testrun has been already run' 

    # create item
    item_uuid = uuid.uuid4()
    testrun_call = models.TestRunCall(
        id=item_uuid.bytes,
        creator=user,
        testrun=testrun,
    )
    session.add(testrun_call)
    session.commit()

    # get items from db
    items = models.TestRunCall.query.all()
    assert len(items) == 1, f'There are {len(items)} TestRun Calls'
    
    # assert properties 
    item = items[0]
    assert item.id == item_uuid.bytes, 'Item ID does not match'
    assert item.testrun_id == testrun.id, 'Item testrun does not match'
    assert item.creator_id == user.id, 'Item createor does not match'
    assert type(item.created) == type(datetime.now())
    assert not item.is_failed, "Default value of 'is_failed' attribute should be 'False'"
    assert item.error_message is None, 'Item error_message attribute is NOT NULL'
    assert not item.is_finished, "Default value of 'is_finished' attribute should be 'False'"
    assert item.testrecords is None, 'Item testrecords attribute is NOT NULL'
    assert item.successful is None, 'Item successful attribute is NOT NULL'
    assert item.error is None, 'Item error attribute is NOT NULL'
    assert item.paused is None, 'Item paused attribute is NOT NULL'
    assert item.duration is None, 'Item duration attribute is NOT NULL'
    assert item.stage is None, 'Item stage attribute is NOT NULL'

    # assert back references
    assert user.testruns_called == [item]
    assert testrun.calls == [item]

    # assert methods
    assert str(item) == str(item_uuid), f'__str__ returns: {str(item)}. Should be item UUID'


@pytest.mark.parametrize('count', [(1,1)])
def test_testrun_call_update(session, test_user, populate):
   # get user and testrun
    user = models.User.query.filter_by(username=test_user.get('username')).first()
    testrun = models.Testrun.query.first()

    # create item
    item_uuid = uuid.uuid4()
    testrun_call = models.TestRunCall(
        id=item_uuid.bytes,
        creator=user,
        testrun=testrun,
    )
    session.add(testrun_call)
    session.commit() 

    # update item
    with open(path.join(path_to_data, 'testrun_calls.json'), 'r') as f:
        calls = json.load(f)
    test_item = calls.get('success')[0]
    test_item.update(calls.get('error')[0])
    testrun_call.is_failed = True
    testrun_call.error_message = test_item.get('error_msg')
    testrun_call.is_finished = True
    testrun_call.testrecords = test_item.get('testrecords')
    testrun_call.successful = test_item.get('successful')
    testrun_call.error = test_item.get('error')
    testrun_call.paused = test_item.get('paused')
    testrun_call.duration = test_item.get('duration')
    testrun_call.stage = test_item.get('stage')
    session.commit()

     # get items from db
    item = models.TestRunCall.query.get(item_uuid.bytes)
    assert bool(item), f'Object could not be retrieved by ID'
    
    # assert properties 
    assert item.is_failed, 'Item is_failed attribute does not match'
    assert item.error_message == test_item.get('error_msg'), 'Item error_message attribute does not match'
    assert  item.is_finished, 'Item is_finished attribute does not match'
    assert item.testrecords == test_item.get('testrecords'), 'Item testrecords attribute does not match'
    assert item.successful == test_item.get('successful'), 'Item successful attribute does not match'
    assert item.error == test_item.get('error'), 'Item error attribute does not match'
    assert item.paused == test_item.get('paused'), 'Item paused attribute does not match'
    assert item.duration == test_item.get('duration'), 'Item duration attribute does not match'
    assert item.stage == test_item.get('stage'), 'Item stage attribute does not match'





