##########################################
#                                        #
#    ENSURE baangt-file service is up    #
#                                        #
##########################################

import pytest
from tests.supports import login, path_to_data, get_csrf, get_datafile_url, dump_html
from app import models
from lxml import html
import os
import json
import requests
import uuid


#
# Import Testrun with data
#
def test_import_simple(client, test_user, default_supports, db_supports):
    # set-up
    url = '/testrun/import'
    db_supports(5)
    login(client, test_user)

    # post import
    with open(os.path.join(path_to_data, 'testrun_simple.xlsx'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    # assert no error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert not bool(flash_area[0].xpath('./div[contains(@class, "alert-warning")]')), 'Error message shown for CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_simple.json'), 'r') as f:
        imported_testrun = json.load(f)
    # testrun
    item = models.Testrun.query.filter_by(name=imported_testrun['testrun'].get('name')).first()
    assert bool(item), 'Imported Testrun NOT saved to DB'
    assert item.description == imported_testrun['testrun'].get('description_import')
    # testcase sequence
    assert len(item.testcase_sequences) == 1
    item = item.testcase_sequences[0]
    assert item.name == imported_testrun['testcase_sequence'].get('name')
    assert item.description == imported_testrun['testcase_sequence'].get('description_import')
    assert item.classname.name == imported_testrun['testcase_sequence'].get('classname')
    # datafile
    assert bool(item.datafiles)
    datafile = item.datafiles[0]
    assert datafile.filename == imported_testrun['datafile'].get('name')
    assert datafile.sheet == imported_testrun['datafile'].get('sheet')
    # testcase
    assert len(item.testcases) == 1
    item = item.testcases[0]
    assert item.name == imported_testrun['testcase'].get('name')
    assert item.description == imported_testrun['testcase'].get('description_import')
    assert item.classname.name == imported_testrun['testcase'].get('classname')
    assert item.browser_type.name == imported_testrun['testcase'].get('browser_type')
    assert item.testcase_type.name == imported_testrun['testcase'].get('testcase_type')
    # teststep sequence
    assert len(item.teststep_sequences) == 1
    item = item.teststep_sequences[0]
    assert item.name == imported_testrun['teststep_sequence'].get('name')
    assert item.description == imported_testrun['teststep_sequence'].get('description_import')
    assert item.classname.name == imported_testrun['teststep_sequence'].get('classname')
    # teststeps
    assert len(item.teststeps) == 5
    for teststep, teststep_ref in zip(item.teststeps, imported_testrun.get('teststeps')):
        assert teststep.name == teststep_ref.get('name')
        assert teststep.description == teststep_ref.get('description_import')
        assert (teststep.locator_type.name if teststep.locator_type else teststep.locator_type) == teststep_ref.get('locator_type') 
        assert teststep.locator == teststep_ref.get('locator')
        assert teststep.value == teststep_ref.get('value')
        assert teststep.value2 == teststep_ref.get('value2')
        assert teststep.comparison == teststep_ref.get('comparison')
        assert str(teststep.timeout) == str(teststep_ref.get('timeout'))
        assert teststep.optional == teststep_ref.get('optional')
        assert teststep.release == teststep_ref.get('release')


def test_import_complex(client, test_user, default_supports, db_supports):
    # set-up
    url='/testrun/import'
    db_supports(10)
    login(client, test_user)

    # import testrun
    files = [
        open(os.path.join(path_to_data, 'testrun_complex.xlsx'), 'rb'),
        open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb'),
        open(os.path.join(path_to_data, 'datafile_2.xlsx'), 'rb'),
    ]
    file_data = {
        'csrf_token': get_csrf(client, '/testrun'),
        'file': files[0],
        'dataFile1': files[1],
        'dataFile2': files[2],
    }
    r = client.post(url, data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'POST Import Testrun. Status code: {r.status_code}'
    for f in files:
        f.close()
    # assert no error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert not bool(flash_area[0].xpath('./div[contains(@class, "alert-warning")]')), 'Error message shown for CORRECT import'

    # assert imported testrun
    with open(os.path.join(path_to_data, 'testrun_complex.json'), 'r') as f:
        tr_ref = json.load(f).get('testrun')
    # testrun
    tr = models.Testrun.query.filter_by(name=tr_ref.get('name')).first()
    assert bool(tr), 'Imported Testrun NOT saved to DB'
    assert tr.description == tr_ref.get('description_import')
    # testcase sequence
    assert len(tr.testcase_sequences) == len(tr_ref.get('testcase_sequences'))
    for tcs, tcs_ref in zip(tr.testcase_sequences, tr_ref.get('testcase_sequences')):
        assert tcs.name == tcs_ref.get('name')
        assert tcs.description == tcs_ref.get('description_import')
        assert tcs.classname.name == tcs_ref.get('classname')
        # datafile
        assert bool(tcs.datafiles)
        datafile = tcs.datafiles[0]
        assert datafile.filename == tcs_ref['datafile'].get('name')
        assert datafile.sheet == tcs_ref['datafile'].get('sheet')
        # testcase
        assert len(tcs.testcases) == len(tcs_ref.get('testcases'))
        for tc, tc_ref in zip(tcs.testcases, tcs_ref.get('testcases')):
            assert tc.name == tc_ref.get('name')
            assert tc.description == tc_ref.get('description_import')
            assert tc.classname.name == tc_ref.get('classname')
            assert tc.browser_type.name == tc_ref.get('browser_type')
            assert tc.testcase_type.name == tc_ref.get('testcase_type')
            # teststep sequence
            assert len(tc.teststep_sequences) == len(tc_ref.get('teststep_sequences'))
            for tss, tss_ref in zip(tc.teststep_sequences, tc_ref.get('teststep_sequences')):
                assert tss.name == tss_ref.get('name')
                assert tss.description == tss_ref.get('description_import')
                assert tss.classname.name == tss_ref.get('classname')
                # teststeps
                assert len(tss.teststeps) == len(tss_ref.get('teststeps'))
                for ts, ts_ref in zip(tss.teststeps, tss_ref.get('teststeps')):
                    assert ts.name == ts_ref.get('name')
                    assert ts.description == ts_ref.get('description_import')
                    assert (ts.locator_type.name if ts.locator_type else ts.locator_type) == ts_ref.get('locator_type') 
                    assert ts.locator == ts_ref.get('locator')
                    assert ts.value == ts_ref.get('value')
                    assert ts.value2 == ts_ref.get('value2')
                    assert ts.comparison == ts_ref.get('comparison')
                    assert str(ts.timeout) == str(ts_ref.get('timeout'))
                    assert ts.optional == ts_ref.get('optional')
                    assert ts.release == ts_ref.get('release')


#
# Import Testrun: Datafile upload
#
def test_datafile_import_simple(application, test_user, default_supports, db_supports):
    # set-up
    url = '/testrun/import'
    db_supports(5)
    client = application.test_client()
    login(client, test_user)

    # import testrun
    tetsrun_name = 'testrun_simple.xlsx'
    with open(os.path.join(path_to_data, tetsrun_name), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, '/testrun'),
            'file': f,
        }
        r = client.post(url, data=file_data)
    assert r.status_code == 302, f'POST Import Testrun. Status code: {r.status_code}'

    # assert uploaded datafile
    # get imported testrun
    item = models.Testrun.query.filter_by(name=tetsrun_name).first()
    datafile_url = '/'.join((
        'http:/',
        application.config.get('BAANGT_DATAFILE_HOST'),
        application.config.get('BAANGT_DATAFILE_GET'),
        item.testcase_sequences[0].datafiles[0].uuid,
    ))
    r = requests.get(datafile_url)
    assert r.status_code == 200

def test_datafile_import_complex(application, test_user, default_supports, db_supports):
    # set-up
    url='/testrun/import'
    db_supports(10)
    client = application.test_client()
    login(client, test_user)

    # import testrun
    tetsrun_name = 'testrun_complex.xlsx'
    files = [
        open(os.path.join(path_to_data, tetsrun_name), 'rb'),
        open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb'),
        open(os.path.join(path_to_data, 'datafile_2.xlsx'), 'rb'),
    ]
    file_data = {
        'csrf_token': get_csrf(client, '/testrun'),
        'file': files[0],
        'dataFile1': files[1],
        'dataFile2': files[2],
    }
    r = client.post(url, data=file_data)
    for f in files:
        f.close()
    assert r.status_code == 302, f'POST Import Testrun. Status code: {r.status_code}'

    # assert uploaded datafiles
    # get imported testrun
    item = models.Testrun.query.filter_by(name=tetsrun_name).first()
    for tcs_index, tcs in enumerate(item.testcase_sequences):
        datafile_url = '/'.join((
            'http:/',
            application.config.get('BAANGT_DATAFILE_HOST'),
            application.config.get('BAANGT_DATAFILE_GET'),
            tcs.datafiles[0].uuid,
        ))
        r = requests.get(get_datafile_url(application.config, tcs.datafiles[0].uuid))
        assert r.status_code == 200, f'Unable access datafile of testcasec sequence {tcs_index}'


#
# DataFile upload
#
def test_datafile_upload(application, test_user):
    # set-up
    url = '/datafile/upload'
    client = application.test_client()

    # unauthorized access
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        file_data = {'datafile': f}
        r = client.post(url, data=file_data)
    # assert redirection response
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # login
    login(client, test_user)

    # get url
    r = client.get(url)
    assert r.status_code == 405, 'Method GET should not be accepted'

    # correct upload
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        file_data = {'datafile': f}
        r = client.post(url, data=file_data)
    assert r.status_code == 200
    response = json.loads(r.data)

    # assert uploaded datafile exists
    r = requests.get(get_datafile_url(application.config, response.get('id')))
    assert r.status_code == 200, 'Unable access uploaded datafile'

def test_datafile_upload_error(client, test_user):
    # set-up
    url = '/datafile/upload'
    login(client, test_user)

    # upload null
    file_data = {'datafile': None}
    r = client.post(url, data=file_data)
    # assert redirection response
    assert r.status_code == 400, f'Upload NULL file. Status code: {r.status_code}'
    response = json.loads(r.data)  
    assert response.get('error')


#
# DataFile Update
#
def test_datafile_update(application, test_user, default_supports, db_supports):
    # set-up
    urls = {
        'datafile_update': '/datafile/update',
        'testrun_import': '/testrun/import',
        'testrun': '/testrun',
        'testcase_sequence': '/testcase_sequence',
    }
    client = application.test_client()
    db_supports(5)

    # unauthorized access
    r = client.post(urls.get('datafile_update'))
    # assert redirection response
    assert r.status_code == 302, f'Unauthorized access. Status code: {r.status_code}'
    page = html.fromstring(r.data)  
    assert '/login?' in page.xpath('//a')[0].attrib.get('href'), 'Unauthorized access: wrong redirection'

    # login
    login(client, test_user)

    # get url
    r = client.get(urls.get('datafile_update'))
    assert r.status_code == 405, 'Method GET should not be accepted'

    # assert form
    r = client.get(urls.get('testcase_sequence'))
    assert r.status_code == 200, 'Unable access TestCaseSequences page'
    # assert 
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages is absent'
    assert not bool(flash_area[0].xpath('./div[contains(@class, "alert-warning")]')), 'Error message shown for CORRECT import'

    # import testrun
    tetsrun_name = 'testrun_simple.xlsx'
    with open(os.path.join(path_to_data, tetsrun_name), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, urls.get('testrun')),
            'file': f,
        }
        r = client.post(urls.get('testrun_import'), data=file_data)
    assert r.status_code == 302, f'POST Import Testrun. Status code: {r.status_code}'
    
    # get testcase sequence of the imported testrun
    item = models.Testrun.query.filter_by(name=tetsrun_name).first().testcase_sequences[0]

    # datafile update form
    r = client.get(urls.get('testcase_sequence'))
    assert r.status_code == 200, 'Unable access TestCaseSequences page'
    # assert form elements
    page = html.fromstring(r.data)
    form = page.xpath('//form[@id="datafileUpdateForm"]')
    assert bool(form), f'Datafile Update Form is absent'
    form = form[0]

    # assert form fields
    # CSFR
    field = form.xpath('.//input[@id="csrf_token"]')
    assert bool(field), 'CSRF field is absent'
    assert field[0].attrib.get('type') == 'hidden', 'CSRF field is NOT hidden'
    # datafile field
    field = form.xpath('.//input[@id="datafile"]')
    assert bool(field), 'Data File field is absent'
    value = bool(field[0].attrib.get('value'))
    assert not value, f'Name field is NOT empty'
    # datafile id
    field = form.xpath('.//input[@id="datafile_id"]')
    assert bool(field), 'Data File ID field is absent'
    assert field[0].attrib.get('type') == 'hidden', 'Data File ID field is NOT hidden'
    # testcase sequence id
    field = form.xpath('.//input[@id="testcaseseq_id"]')
    assert bool(field), 'TestCase Sequence ID field is absent'
    assert field[0].attrib.get('type') == 'hidden', 'TestCAse Sequence ID field is NOT hidden'
    
    # assert submit button
    assert bool(form.xpath('.//*[@type="submit"]'))

    # update datafile
    with open(os.path.join(path_to_data, 'datafile_1.xlsx'), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, urls.get('testcase_sequence')),
            'datafile': f,
            'datafile_id': item.datafiles[0].uuid,
            'testcaseseq_id': item.uuid,
        }
        r = client.post(urls.get('datafile_update'), data=file_data)
    # assert redirection response
    assert r.status_code == 302, f'Update DataFile. Status code: {r.status_code}'
    page = html.fromstring(r.data)  
    assert f'/testcase_sequence?show={file_data.get("testcaseseq_id")}' in page.xpath('//a')[0].attrib.get('href'), \
        'Update DataFile: wrong redirection'

    # assert updatetd datafile exists
    r = requests.get(get_datafile_url(application.config, file_data.get('datafile_id')))
    assert r.status_code == 200, 'Unable access updated datafile'

def test_datafile_update_no_file(client, test_user, default_supports, db_supports):
    # set-up
    urls = {
        'datafile_update': '/datafile/update',
        'testrun_import': '/testrun/import',
        'testrun': '/testrun',
        'testcase_sequence': '/testcase_sequence',
    }
    db_supports(5)
    login(client, test_user)

    # import testrun
    tetsrun_name = 'testrun_simple.xlsx'
    with open(os.path.join(path_to_data, tetsrun_name), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, urls.get('testrun')),
            'file': f,
        }
        r = client.post(urls.get('testrun_import'), data=file_data)
    assert r.status_code == 302, f'POST Import Testrun. Status code: {r.status_code}'
    
    # get testcase sequence of the imported testrun
    item = models.Testrun.query.filter_by(name=tetsrun_name).first().testcase_sequences[0]

    # update datafile with NULL
    file_data = {
        'csrf_token': get_csrf(client, urls.get('testcase_sequence')),
        'datafile': None,
        'datafile_id': item.datafiles[0].uuid,
        'testcaseseq_id': item.uuid,
    }
    r = client.post(urls.get('datafile_update'), data=file_data, follow_redirects=True)
    assert r.status_code == 200, f'Update DataFile with NULL. Status code: {r.status_code}'
    dump_html(r)
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'  
    assert bool(flash_area[0].xpath('./div[contains(@class, "alert-danger")]')), 'No error message for Update DataFile with NULL'


#
# Upload Fake DataFile
#
def test_datafile_upload_fake(client, test_user):
    # set-up
    url = '/datafile/upload'
    login(client, test_user)

    # uploade fake datafile
    datafile = 'datafile_fake.xlsx'
    with open(os.path.join(path_to_data, datafile), 'rb') as f:
        file_data = {'datafile': f}
        r = client.post(url, data=file_data)
    assert r.status_code == 400
    # assert response
    response = json.loads(r.data)
    assert 'error' in response, 'Response does NOT contain ERROR key'
    assert datafile in response.get('error'), 'Error messagge does NOT mention DataFile name'



@pytest.mark.parametrize('count', [(1,1)])#, (5,3)])
def test_datafile_update_fake(client, test_user, populate):
    # set-up
    urls = {
        'datafile_update': '/datafile/update',
        'testcase_sequence': '/testcase_sequence',
    }
    login(client, test_user)

    # get testcase sequence
    item = models.TestCaseSequence.query.first()

    # update with fake datafile
    datafile = 'datafile_fake.xlsx'
    with open(os.path.join(path_to_data, datafile), 'rb') as f:
        file_data = {
            'csrf_token': get_csrf(client, urls.get('testcase_sequence')),
            'datafile': f,
            'datafile_id': item.datafiles[0].uuid,
            'testcaseseq_id': item.uuid,
        }
        r = client.post(urls.get('datafile_update'), data=file_data, follow_redirects=True)

    # assert redirection response
    assert r.status_code == 200, f'Update fake DataFile. Status code: {r.status_code}'
    # assert error message
    page = html.fromstring(r.data)
    flash_area = page.xpath('//div[@id="flash-area"]')
    assert bool(flash_area), 'Flash Messages absent'
    msg = flash_area[0].xpath('./div[contains(@class, "alert-danger")]')
    assert bool(msg), 'Error message is absent'
    assert datafile in msg[0].text, 'Error message does NOT contain exported file name'

